!-----------------------------------------------------------------------!
!                                                                       !
!                               M N e                                   !
!                                                                       !
!-----------------------------------------------------------------------!
!       MNE uses maximum likelihood and moment methods to estimate      !
!   the effective size and migration rate of a metapopulation from      !
!   temporal data on gene frequencies. As a special case, it can also   !
!   estimate the effective size for a single isolated population only.  !
!   For more details, see WANG & WHITLOCK (2003, Genetics) and          !
!   Wang (2001, Genet. Res).                                            !
!                                                                       !
!   Notes:                                                              !
!   1. The maximum Ne allowed can be increased to 2 folds if the        !
!     variables TRMX, TRMX2, COL_LOG in the module is defined "SP"      !
!   2. The codes are compiled with OPENMP in Intel's compiler. For other!
!     compilers without openmp, the few openmp lines must be deleted    !
!   3. Compile command line:                                            !
!      Compile command for optimal performance:                         !
!                                                                       !
! mpiifort *.f90 /fp:source /O2 /Qip /Qipo /Qinline-factor=100 /Qunroll-aggressive /Qopenmp /link /STACK:6400000
!      For Profile-guided Optimizations, use                            !
!        /Qprof-gen /Qprof-use sequentially                             !
!                                                                       !
!      Compile for (Windows): Diagnostics                               !
! ifort *.f90 /fp:source /traceback /check:noarg_temp_created /check:all /Od /link /STACK:6400000
!                                                                       !
!      Linux best performance                                           !
! mpiifort -o mne5 -O2 -static-libgcc -fp-model=precise -ipo -inline-factor=100 -unroll-aggressive -qopenmp -qopenmp-link static mne5.f90
!                                                                       !
!   Update Records:                                                     !
!   Date        Changes                                                 !
!   13/10/04    Discarding alleles with no copies in all focal samples  !
!   28/01/05    Fixed a bug in calculating PBDIS, L703-709              !
!   28/04/05    Refixed the bug in calculating PBDIS, L697-721          !
!   15/11/21    Read genotype data directly                             !
!                                                                       !
!       If you detect any bugs, please report to:                       !
!   Dr. Jinliang Wang                                                   !
!   Institute of Zoology                                                !
!   Regent's Park, London NW1 4RY, UK                                   !
!   Tel: +44 20 74496620  Email: jinliang.wang@ioz.ac.uk                !
!-----------------------------------------------------------------------!
MODULE Module_Data
  INTEGER, PARAMETER :: I4B = SELECTED_INT_KIND(9) ! -2147483648~2147483647
  INTEGER, PARAMETER :: I2B = SELECTED_INT_KIND(4) ! -32768~32767
  INTEGER, PARAMETER :: I1B = SELECTED_INT_KIND(2) ! -128~127
  INTEGER, PARAMETER :: SP = KIND(1.0)
  INTEGER, PARAMETER :: DP = SELECTED_REAL_KIND(2 * PRECISION(1._SP))
  Character(Len = 10), Parameter :: Build = '2022.06.30'
  Character(Len = 7), Parameter :: Version = '2.1.0.0'
  CHARACTER(Len = 12) :: DATE, TIME
  INTEGER :: NLOCI, NTLOCI, NSAMB, MAXLC, NE20, M_ESTIMATE, &
    I_Equilibrium, NPARA(0 : 10000, 2), NFUNC, NINTPA, &
    INIT_POINT, MONITOR, NumThread, Loci_TT, MaximumNe, &
    DataForm, OutFreqEstQ, NPoly, NMono
  INTEGER :: NumMPI, ThreadID, MPIRUN
  REAL(DP) :: PREC, PREC_LOG, FPARA(0 : 10000), Est_4Nm
  CHARACTER(LEN = :), Allocatable :: GtypeFile, ParaFile, &
    OutFile, NewLn
  INTEGER, ALLOCATABLE :: NSAMP(:, :), NGEN(:), NS2(:, :), &
    IBPPAA(:, :), KALLELE(:), IB(:, :), AllAllele(:), &
    NSAMP_T(:, :), NumIndiv(:)
  REAL(DP), ALLOCATABLE :: TRMX(:), TRMX2(:), COL_LOG(:, :)
  REAL(DP), ALLOCATABLE :: FACTOR(:), PPAA(:, :), FAC(:), &
    YFC_TEM(:, :), PA0(:)
  LOGICAL :: Thd0Q
  CHARACTER(Len = 1) :: C09(0 : 9)
  DATA C09/'0','1','2','3','4','5','6','7','8','9'/
CONTAINS 
  PURE FUNCTION ItoC(I)
    INTEGER, INTENT(IN) :: I
    CHARACTER(Len = 10) :: C
    CHARACTER(Len = :), Allocatable :: ItoC
    INTEGER :: J, K, N
    J = Merge(I, -I, I > -1)
    N = 10
    IF(J < 10) THEN
      C(N : N) = C09(J)
    ELSE
      DO
        K = Mod(J, 10)
        J = J / 10
        C(N : N) = C09(K)
        N = N - 1
        IF(J < 10) EXIT
      END DO
      C(N : N) = C09(J)
    END IF
    IF(I >= 0) THEN
      ItoC = C(N : 10)
    ELSE
      ItoC = '-' // C(N : 10)
    END IF
  END FUNCTION ItoC
!
  FUNCTION RtoC(R, C)
    REAL(DP), INTENT(IN) :: R
    CHARACTER(*) :: C
    CHARACTER(Len = 28) :: D
    CHARACTER(Len = :), Allocatable :: RtoC
    WRITE(D, Trim(C)) R
    RtoC = Trim(AdjustL(D))
  END FUNCTION RtoC
!
  SUBROUTINE READ_IN_DATA
    INTEGER :: NTL, NAIndiv
    INTEGER(I4B), ALLOCATABLE :: NA(:)
    INTEGER(I2B), ALLOCATABLE :: Gtype(:, :, :), Allele(:)
    INTEGER(I1B), Allocatable :: AlleleID(:)
    CHARACTER(LEN = 9), ALLOCATABLE :: Str(:), GStr(:)
    CHARACTER(LEN = 999) :: C999
    Character(Len = :), Allocatable :: StrLine
!$ INTEGER :: OMP_GET_NUM_PROCS
    REAL(DP) :: XY(2)
    LOGICAL :: ExistQ
    INTEGER(I4B) :: J
    REAL(DP), EXTERNAL :: gammln
!Read parameters
    IF(Thd0Q) THEN
      WRITE(*, '(/99A)')('-', J = 1, 10) 
      WRITE(*, '(A)')'Parameters'
      WRITE(*, '(99A)')('-', J = 1, 10)
    END IF
    OPEN(UNIT = 10, FILE = ParaFile)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) M_ESTIMATE
    IF(Thd0Q) WRITE(*, '(A,T35,A)')'Model : ', &
      Merge('Migration', 'Isolation', M_ESTIMATE == 1)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) I_Equilibrium
    IF(I_Equilibrium /= 0 .AND. I_Equilibrium /= 1) THEN
      IF(Thd0Q) WRITE(*, '(A)') 'The indicator for ' // & 
        'drift-migration equilibrium should be 0/1!', &
        'Program stops!'
      STOP
    END IF
    IF(Thd0Q .AND. M_ESTIMATE == 1) WRITE(*, '(A,T35,A)') &
      'Equilibrium : ', Merge('Yes', 'No ', &
      I_Equilibrium == 1)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) MaximumNe
    IF(Thd0Q) WRITE(*, '(A,T35,A)') &
      'Maximal Ne : ', ItoC(MaximumNe)
    NE20 = 2 * MaximumNe
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) Monitor
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) NumThread
!$  IF(NumThread <= 0) NumThread = OMP_GET_NUM_PROCS()
    IF(Thd0Q) WRITE(*, '(A,T35,A)') &
      '#OpenMP threads/(MPI process) : ', ItoC(NumThread), &
      '#MPI processes : ', ItoC(NumMPI)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) NTL       !#All_LOCI
    Loci_TT = NTL
    IF(Thd0Q) WRITE(*, '(A,T35,A)') '#Loci : ', ItoC(NTL)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) C999      !data file
    GtypeFile = Trim(AdjustL(C999))
    Inquire(FILE = GtypeFile, Exist = ExistQ)
    IF(.NOT. ExistQ) THEN
      IF(Thd0Q) WRITE(*, '(A)') 'Genotype file, ' // &
        GtypeFile // ', does not exist!', 'Program stops!'
      STOP
    END IF 
    IF(Thd0Q) WRITE(*, '(A,T35,A)') 'Data file : ', GtypeFile
    J = Index(GtypeFile, '.', BACK = .TRUE.)
    IF(J > 0) THEN
      OutFile = GtypeFile(1 : J) // 'mNe'
    ELSE
      OutFile = GtypeFile // '.mNe'
    END IF
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) DataForm  !genotype form
    IF(DataForm < 0 .OR. DataForm > 2) THEN
      WRITE(*, *)'Errors in DataForm. ' // &
      'It is out of range [0,1,2] for ' // &
      '{GenePop, 0/1/2/3, 2 alleles/genotype]' // &
      NewLn // ' Please check your DATA ' // &
      'in file ' // ParaFile // ' and re-run.'
      STOP
    END IF
    IF(Thd0Q) WRITE(*, '(A)') &
      'Genotype format [0:GenePop, 1:0/1/2/3, ' // &
      '2:2Alleles/genotype]: ' // ItoC(DataForm)
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) NSAMB
    IF(Thd0Q) WRITE(*, '(A,T35,A)') &
      '#Focal samples : ', ItoC(NSAMB)
    ALLOCATE(NGEN(NSAMB), NumIndiv(0 : NSAMB))
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) &
      (NGEN(I), I = 1, NSAMB)
    DO I = 1, Merge(NSAMB, 0, Thd0Q)
      WRITE(*, '(A,T35,A)') '#Generations of focal sample ' // &
       ItoC(I) // ' : ', ItoC(NGEN(I))
    END DO
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) &
      (NumIndiv(I), I = 1, NSAMB)
    DO I = 1, Merge(NSAMB, 0, Thd0Q)
      WRITE(*, '(A,T35,A)') '#Individuals of focal sample ' // &
       ItoC(I) // ' : ', ItoC(NumIndiv(I))
    END DO  
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) NumIndiv(0)
    IF(M_ESTIMATE == 0) NumIndiv(0) = 0
    IF(Thd0Q .AND. M_ESTIMATE == 1) WRITE(*, '(A,T35,A)') &
      '#Individuals source sample : ', ItoC(NumIndiv(0))
    READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) INIT_POINT
    IF(M_ESTIMATE == 0) INIT_POINT = 1
    IF(Thd0Q .AND. M_ESTIMATE == 1) WRITE(*, '(A,T35,A)') &
      '#Initial points : ', ItoC(INIT_POINT)
    IF(Thd0Q) THEN
      WRITE(*, '(//)')
      WRITE(*, '(99A)')('-', J = 1, 10)
      WRITE(*, '(A)') ' Analysis '
      WRITE(*, '(99A)')('-', J = 1, 10)
    END IF
    CLOSE(10)
!Read genotype data    
    NAIndiv = Sum(NumIndiv(0 : NSAMB))
    IF(Thd0Q) WRITE(*, '(/A)') &
      'Reading genotype data (' // ItoC(NAIndiv) // &
      ' individuals & ' // ItoC(NTL) // ' loci)...'
    Allocate(Gtype(2, NTL, NAIndiv), Str(1))
    OPEN(10, File = GtypeFile)
    NN = 0
    IF(DataForm == 0) THEN  !GenePop Format
      Allocate(GStr(NTL))
      DO K = 1, NSAMB + 1
        IF(K <= NSAMB) THEN
          L = Sum(NumIndiv(0 : K - 1))
          N = L + NumIndiv(K)
        ELSE
          L = 0
          N = NumIndiv(0)
        END IF
        DO N = L + 1, N
          NN = NN + 1
          IF(Thd0Q .AND. Mod(NN, 100) == 0) &
            WRITE(*, '(A,$)') Char(13) // '  Indiv:' // ItoC(NN)
          READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) &
            Str(1), (GStr(J), J = 1, NTL)
          DO J = 1, NTL
            GStr(J) = Trim(AdjustL(GStr(J)))
            M = Len_Trim(GStr(J)) / 2
            READ(GStr(J)(1 : M), *, IOSTAT = I) Gtype(1, J, N)
            IF(I == 0) READ(GStr(J)(1 + M : M + M), *, &
                            IOSTAT = I) Gtype(2, J, N)
            IF(I /= 0) THEN
              WRITE(*, '(//A)') 'Error reading genotype at locus ' // &
                ItoC(J) // ' of individual ' // ItoC(N) // '.'
              STOP
            END IF
          END DO
        END DO
      END DO
    ELSE IF(DataForm == 1) THEN  !0/1/2/3 format
      ALLOCATE(Character(Len = NTL + 1) :: StrLine)
      DO K = 1, NSAMB + 1          
        IF(K <= NSAMB) THEN
          L = Sum(NumIndiv(0 : K - 1))
          N = L + NumIndiv(K)
        ELSE
          L = 0
          N = NumIndiv(0)
        END IF
        DO N = L + 1, N
          NN = NN + 1
          IF(Thd0Q .AND. Mod(NN, 100) == 0) &
            WRITE(*, '(A,$)') Char(13) // '  Indiv:' // ItoC(NN)
          READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) &
            Str(1), StrLine
          J = Len_Trim(StrLine)
          IF(J /= NTL) THEN
            WRITE(*, '(//A)', Advance = 'NO') &
              'Error reading genotype data ' // &
              'of individual ' // ItoC(N) // '.' // NewLn // &
              'The number of genotypes should be ' // ItoC(NTL) &
              // ', but is found to be ' 
            IF(J < NTL) THEN
              WRITE(*, '(A)') ItoC(J) // '!'  
            ELSE
              WRITE(*, '(A)') '>= ' // ItoC(J) // '!'
            END IF
            STOP
          END IF
          DO J = 1, NTL
            IF(StrLine(J : J) == '0') THEN
              Gtype(1 : 2, J, N) = 2_I2B
            ELSE IF(StrLine(J : J) == '1') THEN
              Gtype(1 : 2, J, N) = (/1_I2B, 2_I2B/)
            ELSE IF(StrLine(J : J) == '2') THEN
              Gtype(1 : 2, J, N) = 1_I2B
            ELSE
              Gtype(1 : 2, J, N) = 0_I2B
            END IF
          END DO
        END DO
      END DO
    ELSE
      DO K = 1, NSAMB + 1
        IF(K <= NSAMB) THEN
          L = Sum(NumIndiv(0 : K - 1))
          N = L + NumIndiv(K)
        ELSE
          L = 0
          N = NumIndiv(0)
        END IF
        DO N = L + 1, N
          NN = NN + 1
          IF(Thd0Q .AND. Mod(NN, 100) == 0) &
            WRITE(*, '(A,$)') Char(13) // '  Indiv:' // ItoC(NN)
          READ(10, *, IOSTAT = IOS, ERR = 9, End = 9) &
            Str(1), Gtype(1:2, 1 : NTL, N)
        END DO
      END DO
    END IF
    CLOSE(10)
    IF(Thd0Q) WRITE(*, '(A)') Char(13) // &
      '  Done for indiv = 1 to ' // ItoC(NN)
!Summerize genotype data
!#alleles at each locus, counts of alleles of each sample
    IF(Thd0Q) WRITE(*, '(/A, $)') &
      'Preprocessing of genotype data...'
    ALLOCATE(NA(NTL), Allele(999), AlleleID(9999))
    NLOCI = 0
    AlleleID(:) = 0_I1B
    DO L = 1, NTL
      N = 0
      DO I = 1, NAIndiv
        DO J = 1, 2
          IF(Gtype(J, L, I) < 1) CYCLE  !Missing
          IF(AlleleID(Gtype(J, L, I)) > 0_I1B) CYCLE
          N = N + 1
          Allele(N) = Gtype(J, L, I)
          AlleleID(Gtype(J, L, I)) = 1_I1B
        END DO
      END DO
      NA(L) = N
      FORALL(J = 1 : N) AlleleID(Allele(J)) = 0_I1B
      IF(N > 1) NLOCI = NLOCI + 1
    END DO
    ALLOCATE(KALLELE(1 : NLOCI))
    N = 0
    DO L = 1, NTL
      IF(NA(L) < 2) CYCLE
      N = N + 1
      KALLELE(N) = NA(L)
    END DO
    NPoly = N
    NMono = NTL - NPoly
    MAXLC = SUM(KALLELE)
    ALLOCATE(NS2(MAXLC, 0 : NSAMB), &
           NSAMP(MAXLC, 0 : NSAMB), &
           AllAllele(MAXLC), &
           NSAMP_T(MAXLC, 1 - M_ESTIMATE : NSAMB))
    L1 = 0
    AlleleID(:) = 0_I1B
    DO L = 1, NTL
      IF(NA(L) < 2) CYCLE
      N = 0
      DO I = 1, NAIndiv
        DO J = 1, 2
          IF(Gtype(J, L, I) < 1) CYCLE  !Missing
          IF(AlleleID(Gtype(J, L, I)) > 0_I1B) CYCLE
          N = N + 1
          Allele(N) = Gtype(J, L, I)
          AlleleID(Gtype(J, L, I)) = Int(N, I1B)
        END DO
        IF(N == NA(L)) EXIT
      END DO
      AllAllele(L1 + 1 : L1 + N) = Allele(1 : N)
      DO I = Merge(0, 1, NumIndiv(0) > 0), NSAMB
        IF(I == 0) THEN
          L2 = 0
        ELSE
          L2 = Sum(NumIndiv(0 : I - 1))
        END IF
        NSAMP(L1 + 1 : L1 + N, I) = 0
        DO J = L2 + 1, L2 + NumIndiv(I)
          IF(Gtype(1, L, J) > 0)THEN
            M = AlleleID(Gtype(1, L, J))
            NSAMP(L1 + M, I) = NSAMP(L1 + M, I) + 1
          END IF
          IF(Gtype(2, L, J) > 0)THEN
            M = AlleleID(Gtype(2, L, J))
            NSAMP(L1 + M, I) = NSAMP(L1 + M, I) + 1
          END IF
        END DO
      END DO
      FORALL(J = 1 : N) AlleleID(Allele(J)) = 0_I1B
      L1 = L1 + N
    END DO
    DO I = Merge(0, 1, M_ESTIMATE == 1), NSAMB
      N = 0
      DO J = 1, NLOCI
        N2 = SUM(NSAMP(N + 1 : N + KALLELE(J), I))
        NS2(N + 1 : N + KALLELE(J), I) = N2
        N = N + KALLELE(J)
      END DO
    END DO
    CLOSE(10)
    FORALL(I = 1 : MAXLC, J = 1 - M_ESTIMATE : NSAMB) &
      NSAMP_T(I, J) = NSAMP(I, J)
!
    KK = SUM(KALLELE(1 : NLOCI))
    JJ = NLOCI
    N = KK + 1
    NewLocus = 0
    M1 = 1 - M_ESTIMATE
    DO J = NLOCI, 1, -1
      NN = 0
      DO K = KALLELE(J), 1, -1
        N = N - 1
        IF(ALL(NSAMP(N, M1 : NSAMB) == 0) .OR. &
           ALL(NSAMP(N, M1 : NSAMB) == NS2(N, M1 : NSAMB))) THEN
          DO I = M1, NSAMB
!$OMP SIMD
            DO L1 = N + 1, KK
              NSAMP(L1 - 1, I) = NSAMP(L1, I)
              NS2(L1 - 1, I) = NS2(L1, I)
            END DO
          END DO
          KALLELE(J) = KALLELE(J) - 1
          IF(NewLocus > 0) NewLocus = NewLocus - 1
        ELSE IF(M_ESTIMATE == 1 .AND. &
            ALL(NSAMP(N, 1 : NSAMB) == 0)) THEN
          NN = NN + 1
          IF(NewLocus == 0) THEN
            NewLocus = N
          ELSE
            NSAMP(NewLocus, 0) = NSAMP(NewLocus, 0) + NSAMP(N, 0)
            IF(NN == 1) NS2(NewLocus, 0 : NSAMB) = &
              NS2(NewLocus, 0 : NSAMB) + NS2(N, 0 : NSAMB)
            DO I = M1, NSAMB
!$OMP SIMD
              DO L1 = N + 1, KK
                NSAMP(L1 - 1, I) = NSAMP(L1, I)
                NS2(L1 - 1, I) = NS2(L1, I)
              END DO
            END DO
            KALLELE(J) = KALLELE(J) - 1
            NewLocus = NewLocus - 1
          END IF
        END IF
      END DO
      IF(KALLELE(J) == 0)THEN
        DO K = J + 1, JJ
          KALLELE(K - 1) = KALLELE(K)
        END DO
        NLOCI = NLOCI - 1
      END IF
    END DO
    IF(Thd0Q) WRITE(*, '(A)') ' Done.'
!Initialising variables
    IF(Thd0Q) WRITE(*, '(/A)') 'Initialising variables...'
    PREC = 1.E-10_DP
    PREC_LOG = LOG(PREC)
    NINTPA = 100
    J = Max(Maxval(NS2( : MAXLC, 1 - M_ESTIMATE : NSAMB)), NE20)
    ALLOCATE(fac(0 : J))
    XY(2) = 0.D0
    DO I = 1 - M_ESTIMATE, NSAMB
      XY(2) = XY(2) + 1.D0 / NumIndiv(I)
    END DO
    XY(2) = (NSAMB + M_ESTIMATE) / XY(2) !Harmonic sample size
    XY(1) = 1.D0 + 0.025D0 / ((XY(2) - XY(2) ** 0.75) * &
      (NGEN(NSAMB) - 0.8) ** 0.5) + 0.0000005D0
    DO i = 0, J
      fac(i) = gammln(XY(1) + I)
    END DO
    XY(1) = 1._DP / NE20
    XY(2) = Merge(0.1_DP, 0._DP, M_ESTIMATE == 1)
    CALL TrmxDim(NE20, XY(1), XY(2), MAXDIM)
    IF(MAXDIM > 0 .AND. MAXDIM < 200000000)THEN
      ALLOCATE(TRMX(0 : MAXDIM), STAT = KK)
    ELSE
      KK = 1
    END IF
    IF(KK /= 0)THEN
      ISTEP = -MIN(MAX(NINT(NE20 * .02), 2), 500)
      DO I = NE20, 2, ISTEP
        IF(ALLOCATED(TRMX))DEALLOCATE(TRMX)
        XY(1) = 1._DP / I
        CALL TrmxDim(I, XY(1), XY(2), MAXDIM)
        IF(MAXDIM > 0 .AND. MAXDIM < 200000000) &
          ALLOCATE(TRMX(0 : MAXDIM), STAT = KK)
        IF(KK == 0) EXIT
      END DO
      IF(Thd0Q) WRITE(*, '(/A)') 'The maximum ' // &
        'value of Ne set in your data is: ' &
        // ItoC(NE20 / 2) // NewLn // &
        '  which is too large that your computer may have ' // & 
        'not enough RAM to run the program. ' // &
        NewLn // '  The suggested maximum ' // &
        'value of Ne is:' // ItoC(NINT(I * .5)) // &
        NewLn // NewLn // &
        'The program will run with ' // &
        'this maximum value of Ne. However, ' // &
        NewLn // '  if memory problem ' // &
        'still exists in the run, ' // &
        'please set maximum value of Ne' // NewLn // &
        '  in your data input to 90% of the value ' // &
        'suggested, and re-run the program' // NewLn
      CALL Sleeping(4)
      NE20 = I
      MaximumNe = NINT(I * .5)
    END IF
    IF(ALLOCATED(TRMX))DEALLOCATE(TRMX)
    IF(Thd0Q) WRITE(*, '(/A)') 'Initialising variables finished.'
    RETURN
9   IF(Thd0Q) WRITE(*, '(//A)')'Errors in DATA INPUT. ' // &
      'Insufficient data or incorrect' // NewLn // &
      '  format. Please check your DATA ' // &
      'in file ' // ParaFile // ' and re-run.'
    STOP
  END SUBROUTINE READ_IN_DATA
!
  FUNCTION Day_of_week(yyyymmdd) !YearMonthDay
    INTEGER :: i, j, k, mm, yy, dd
    Character(*) :: yyyymmdd
    Character(Len = 9) :: WeekDay(0 : 6)
    CHARACTER(Len = :), Allocatable :: Day_of_week
    DATA WeekDay/'Saturday ', 'Sunday   ', 'Monday   ', &
      'Tuesday  ', 'Wednesday', 'Thursday ', 'Friday   '/
    READ(yyyymmdd(1 : 4), '(I4)') yy
    READ(yyyymmdd(5 : 6), '(I4)') mm
    READ(yyyymmdd(7 : 8), '(I4)') dd
    IF(mm.le.2) THEN
       mm=mm+12
       yy=yy-1
    END IF
    j = yy / 100
    k = MOD(yy, 100)
    i = MOD(dd + ((mm+1)*26)/10 + k + k/4 + j/4 + 5*j, 7)
    Day_of_week = Trim(WeekDay(i))
  END FUNCTION Day_of_week
END MODULE Module_Data
!
PROGRAM MAIN
  Use Module_Data
!$  include 'mpif.h'
  CHARACTER * 3 :: PARANAME(2)
  REAL(SP) :: MLNE2, MLM, MTNE, MTM
  REAL(DP) :: X1, Y1, DY, fmax, fmax2, xmax(2), &
    xmax2(2), VarStep(2), XII(2, 2), &
    XI(2, 2), XIII(2, 2), FUNC, MAX_LD, &
    CI95(-100 : 100, 2, 2), RANGE(2, 2), XA(3), YA(3)
  REAL(DP) :: BRENT, FstFunc
  INTEGER :: BOUND(2, 2)
  REAL(DP), Allocatable :: LocusWt(:), Weight(:, :)
  INTEGER(I4B), Allocatable :: Locus(:)
  Logical :: ExistQ
  CHARACTER(Len = :), Allocatable :: Str
  DATA XII/8._DP, 2 * 0._DP, 0.05_DP/, &
    PARANAME/' Ne', '  m'/, &
    XIII/4._DP, 2 * 0._DP, 0.001_DP/
  EXTERNAL :: BRENT, FstFunc, ShowVersion, Output, &
    GetRunCmmd, Indexx
  Allocate(Character(Len = 1) :: GtypeFile, ParaFile, &
    OutFile, NewLn)
  NewLn = New_Line('a')
  CALL GetRunCmmd
  CALL InitializeThreads
  CALL DATE_AND_TIME(DATE, TIME)
  IF(Thd0Q) CALL ShowVersion(6)
  Inquire(FILE = ParaFile, Exist = ExistQ)
  IF(.NOT. ExistQ) THEN
    IF(Thd0Q) WRITE(*, '(/A)') 'Parameter file, ' // &
      ParaFile // ', does not exist!' // &
      NewLn // 'Program stops!'
    STOP
  END IF
  CALL READ_IN_DATA
!$ IF(NumMPI > 1) CALL MPI_BARRIER(MPI_COMM_WORLD, I)
  ALLOCATE(Character(Len = 1) :: Str)
  ALLOCATE(FACTOR(MAXLC), IBPPAA(MAXLC, 2), &
    YFC_TEM(MAXLC, NSAMB), PA0(MAXLC))
  NPARA(0, 1) = -1
  NTLOCI = SUM(KALLELE(1 : NLOCI))
  IF(Thd0Q) THEN
    WRITE(*, '(/99A)')('-', J = 1, 30)
    WRITE(*, '(A)') ' Analysis by moment estimator'
    WRITE(*, '(99A)')('-', J = 1, 30)
  END IF
  CALL MTEST(MTNE, MTM)
  IF(Thd0Q .AND. MONITOR > 0)THEN
    write(*, '(A)', Advance = 'NO') &
      'Moment estimates: Ne = ' // &
      RtoC(Dble(MTNE), '(F12.2)')
    IF(M_ESTIMATE == 1)THEN
      write(*, '(A)') '  m = ' // RtoC(Dble(MTM), '(F12.4)')
    ELSE
      write(*, '(A)')
    END IF
  END IF
  IF(Thd0Q .AND. MTNE < 0.) WRITE(*, '(A)') &
    '  Ne < 0 indicates an infinitely large estimate of Ne'
  IF(Thd0Q .AND. M_ESTIMATE == 1 .AND. MTM < 0) &
    WRITE(*, '(A)') '  m < 0 indicates NO ' // &
      'valid estimate of migration rate'
  IF(Thd0Q) THEN
    WRITE(*, '(/99A)')('-', J = 1, 34)
    WRITE(*, '(A)') ' Analysis by likelihood estimator'
    WRITE(*, '(99A)')('-', J = 1, 34)
    WRITE(*, '(/A,$)')'Finding duplicated loci...'
  END IF
  NTLOCI = 0
  DO I = 1, NLOCI
    IF(KALLELE(I) == 2)THEN
      FACTOR(NTLOCI + 1) = 1._DP
      FACTOR(NTLOCI + 2) = -1._DP
      NTLOCI = NTLOCI + 2
    ELSE
      X1 = 1._DP - 1._DP / KALLELE(I)
      DO J = 1, KALLELE(I)
        NTLOCI = NTLOCI + 1
        FACTOR(NTLOCI) = X1
      END DO
    END IF
  END DO
  KK = 1 - M_ESTIMATE
  IDUM = -1
  Allocate(LocusWt(NTLOCI), Weight(KK : NSAMB, 4))
  DO I = 1, 4
    DO J = KK, NSAMB
      Weight(J, I) = RAND1(IDUM)
    END DO
  END DO
  Allocate(Locus(NTLOCI))
  J = 0
  DO I = 1, NTLOCI
    IF(NSAMP(I, KK) <= NS2(I, KK) * 0.5) CYCLE
    J = J + 1
    Locus(J) = I
  END DO
  DO K = KK, NSAMB
!$OMP SIMD
    DO I = 1, J
      NSAMP(Locus(I), K) = NS2(Locus(I), K) - NSAMP(Locus(I), K)
    END DO
  END DO
  X1 = 0.5D0 / Sum(NumIndiv(0 : NSAMB))
!$OMP SIMD
  DO I = 1, NTLOCI
    IF(FACTOR(I) < 0._DP) THEN
      LocusWt(I) = -1.D0
    ELSE
      Y1 = 0.D0
      DO J = KK, NSAMB
        Y1 = Y1 + ((0.5D0 + J) / (0.1D0 + Weight(J, 3))) * &
          X1 * NSAMP(I, J) * Weight(J, 1) + ((0.5D0 + J) / &
          (0.1D0 + Weight(J, 4))) * X1 * NS2(I, J) * Weight(J, 2)
      END DO
      LocusWt(I) = Y1
    END IF
  END DO
  CALL indexx(NTLOCI,LocusWt,Locus)
  DO I = 1, NTLOCI
    IF(FACTOR(Locus(I)) >= 0.D0) THEN
      J = I
      EXIT
    END IF
  END DO
  I = Locus(J)
  Y1 = LocusWt(I)
  DO K = J + 1, NTLOCI
    J = Locus(K)
    IF(ABS(LocusWt(J) - Y1) < 1.D-17) THEN
      FACTOR(I) = FACTOR(I) + FACTOR(J)
      FACTOR(J) = -1.D0
    ELSE
      I = J
      Y1 = LocusWt(J)
    END IF
  END DO
  K = 0
  DO I = 1, NTLOCI
    IF(FACTOR(I) < 0._DP) CYCLE
    K = K + 1
    Locus(K) = I
    IF(K /= I) FACTOR(K) = FACTOR(I)    
  END DO
  FACTOR(K + 1 : NTLOCI) = -1.D0
  NTLOCI = K
  Allocate(IB(K, KK : NSAMB))
  DO J = KK, NSAMB
!$OMP SIMD
    DO I = 1, K
      IB(I, J) = NSAMP(Locus(I), J)
    END DO
  END DO
  DeAllocate(NSAMP)
  Allocate(NSAMP(K, KK : NSAMB))
  DO J = KK, NSAMB
!$OMP SIMD
    DO I = 1, K
      NSAMP(I, J) = IB(I, J)
    END DO
  END DO
  DO J = KK, NSAMB
!$OMP SIMD
    DO I = 1, K
      IB(I, J) = NS2(Locus(I), J)
    END DO
  END DO
  DeAllocate(NS2)
  Allocate(NS2(K, KK : NSAMB))
  DO J = KK, NSAMB
!$OMP SIMD
    DO I = 1, K
      NS2(I, J) = IB(I, J)
    END DO
  END DO
  DeAllocate(IB)
  IF(Thd0Q) WRITE(*, '(A)')' Done.'
  IF(M_ESTIMATE == 1)THEN
    IF(Thd0Q) WRITE(*, '(A,$)') 'Generate source ' // &
      'pop allele frequency distribution...'
    NINTPA = NINTPA / 2
!$OMP SIMD
    DO I = 1, NTLOCI
      PA0(I) = NSAMP(I, 0) / DBLE(NS2(I, 0))
    END DO
    DO
      NINTPA = NINTPA * 2
      ALLOCATE(PPAA(0 : NINTPA, NTLOCI))
      CALL GTDIS(NINTPA_FLAG)
      IF(NINTPA_FLAG == 0) EXIT
      DEALLOCATE(PPAA)
    END DO
    IF(Thd0Q) WRITE(*, '(A)') ' Done.'
  END IF
  IF(M_ESTIMATE == 1 .AND. I_Equilibrium /= 0) THEN
! Estimate 4Nm when the 1st sample was 
!    taken at an equilibrium population
    IF(Thd0Q) WRITE(*, '(A)') 'Estimating 4Nm...'
    X1 = BRENT(0.01_DP, 5._DP, 1000._DP, &
               FstFunc, 0.01_DP, Est_4Nm)
    IF(Thd0Q) WRITE(*, '(A, F10.2)') 'Estimated 4Nm=', Est_4Nm
  END IF
  IF(Thd0Q) WRITE(*, '(/A)') 'Likelihood ' // &
    'of valid initial estimates...'
  NFUNC = 0
  LFLAG = 0
  IF(MTNE > 0.5 .AND. MTNE <= NE20 * .5) THEN
    XMAX(1) = MTNE * 2._DP
  ELSEIF(MTNE > NE20 * .5 .AND. MTNE < NE20) THEN
    XMAX(1) = MTNE
  ELSE
    XMAX(1) = RAND1(IDUM) * NE20 * .1 + 1._DP
    LFLAG = 1
  END IF
  IF(MTM <= 0.001) THEN
    XMAX(2) = 0.001
  ELSEIF(MTM > 0.999) THEN
    XMAX(2) = 0.999
  ELSE
    XMAX(2) = MTM
  END IF
  XMAX2 = XMAX
  X1 = FUNC(XMAX, 2, NINT(XMAX(1)))
  IF(X1 > 1.E99_DP) THEN
    II = 0
    DO
      II = II + 1
      IF(II > 2000) THEN
        IF(Thd0Q) WRITE(*, '(/A)') 'Unable to ' // &
          'find valid initial valid Ne/m values ' // &
          'for the iterative ML method!' // &
          NewLn // 'Program stops!' // &
          NewLn // 'For migration model, ' // &
          'changing equilibrium model may help.'
        STOP
      END IF
      IF(Thd0Q .AND. Mod(II, 10) == 0) WRITE(*, '(A)') &
        'Finding initial valid Ne/m values, ' // &
        'Attempt = ' // ItoC(II)
      XMAX(1) = RAND1(IDUM) * NE20 * .1 + 1._DP
      IF(LFLAG > 0)THEN
        LFLAG = LFLAG + 1
        XMAX(2) = XMAX2(2)
      ELSE
        XMAX(2) = RAND1(IDUM) * 0.1
      END IF
      FMAX = FUNC(XMAX, 2, NINT(XMAX(1)))
      IF(FMAX < 1.E99_DP)EXIT
      IF(LFLAG > 1000)LFLAG = 0
    END DO
  END IF
  IF(Thd0Q) WRITE(*, '(/A)') 'Powell quadratic optimization...'
  IF(MTNE > 50000 .OR. MTNE <= 0.0) THEN
    X1 = Merge(100000.0, MTNE, MTNE <= 0.0 .OR. MTNE > 100000.0)
    XII(1, 1) = XII(1, 1) + X1 * 0.2
  ELSE IF(MTNE < 100) THEN
    XII(1, 1) = XII(1, 1) + 4
  ELSE IF(MTNE < 1000) THEN
    XII(1, 1) = XII(1, 1) + 40
  ELSE IF(MTNE < 10000) THEN
    XII(1, 1) = XII(1, 1) + 800
  ELSE
    XII(1, 1) = XII(1, 1) + MTNE * 0.2
  END IF
  XI = XII
  CALL POWELL(xmax, XI, ITER, MAX_LD, M_ESTIMATE, 0, MONITOR)
  MLNE2 = REAL(XMAX(1))
  MLM = REAL(XMAX(2))
!Trying different random starting points
  DO II = 1, INIT_POINT - 1
    DO
      XMAX(1) = RAND1(IDUM) * NE20 * .1 + 1._DP
      XMAX(2) = RAND1(IDUM)
      FMAX = FUNC(XMAX, 2, NINT(XMAX(1)))
      IF(FMAX > 1.E99_DP)CYCLE
      IF(Thd0Q .AND. MONITOR >= 1) &
        WRITE(*, '(A, I4, F12.2, F9.4)') &
        'ML estimate with random starting point', &
        II, (XMAX(J), J = 1, 1 + M_ESTIMATE)
      XI = XII
      CALL POWELL(xmax, XI, ITER, FMAX, M_ESTIMATE, 0, MONITOR)
      IF(FMAX < MAX_LD)THEN
        MAX_LD = FMAX
        MLNE2 = REAL(XMAX(1))
        MLM = REAL(XMAX(2))
      END IF
      IF(Thd0Q .AND. MONITOR >= 1)THEN
        write(*, '(A, F12.2)', Advance = 'NO') &
          '  Estimate: Ne = ', Xmax(1) * .5
        IF(M_ESTIMATE == 1) write(*, '(A, F12.4)', &
          Advance = 'NO') '    m = ', XMAX(2)
        write(*, '(A, F12.4/)') '   Log(ML) = ', -FMAX
      END IF
      EXIT
    END DO
  END DO
  IF(Thd0Q .AND. MONITOR > 0) THEN
    Str = RtoC(Dble(MLNE2 / 2), '(F12.2)')
    K2 = Max(Len(Str), 5)
    IF(M_ESTIMATE >= 1)THEN
      WRITE(*, '(/A' // ItoC(K2) // ',4A15)') &
        'ML_Ne', 'ML_m', 'Log(ML)', 'MT_Ne', 'MT_m'
      WRITE(*, '(F' // ItoC(K2) // &
        '.2, 2F15.4, F15.2, 2F15.4/)') &
        MLNE2 / 2, MLM, -MAX_LD, MTNE, MTM
    ELSE
      WRITE(*, '(' // 'A' // ItoC(K2) // &
        ',2A15)')'ML_Ne', 'Log(ML)', 'MT_Ne'
      WRITE(*, '(F' // ItoC(K2) // &
        '.2, 2F15.4/)')MLNE2 / 2, -MAX_LD, MTNE
    END IF
    IF(MTNE < 0.) WRITE(*, '(A)') '  Ne < 0 indicates ' // &
      'an infinitely large estimate of Ne'
    IF(M_ESTIMATE == 1 .AND. MTM < 0) WRITE(*, '(A)') &
      '  MT_m < 0 indicates NO valid estimate of migration rate'
  END IF
  RANGE(1, 1) = 2._DP
  RANGE(2, 1) = NE20
  RANGE(:, 2) = (/0.001_DP, 1.0_DP/)
  K2 = 0
1 IF(K2 /= 0)THEN
    IF(M_ESTIMATE == 1)THEN
      XI = XIII
      CALL POWELL(xmax, XI, ITER, &
                  MAX_LD, M_ESTIMATE, 0, MONITOR)
    ELSE
      MAX_LD = FMAX
    END IF
    MLNE2 = REAL(XMAX(1))
    MLM = REAL(XMAX(2))
  END IF
  CI95 = 0._DP
  CI95(0, 1, 1) = MLNE2
  CI95(0, 1, 2) = MLM
  CI95(0, 2, :) = MAX_LD
  VarStep(1) = MAX(MLNE2 * .01_DP, 2._DP)
  VarStep(2) = MAX(MLM * 0.05_DP, 0.01_DP)
  BOUND = 0
! Finding 95% confidence interval for each parameter
  DO K = 1, 1 + M_ESTIMATE
    DO K2 = 1, 2
      IF(Thd0Q) WRITE(*, '(/A, A)') &
        'Finding ' // Merge('+', '-', K2 == 1) // &
        '95% confidence interval ' // &
        'for parameter', PARANAME(K)
      IF(K2 == 1)THEN
        XSTEP = REAL(VarStep(K))
        KSTEP = 1
      ELSE
        XSTEP = -REAL(VarStep(K))
        KSTEP = -1
      END IF
      XMAX2(:) = CI95(0, 1, :)
      FMAX2 = CI95(0, 2, K)
      KK = 0
      YSTEP = 0.005
      DO
        XMAX = XMAX2
        XSTEP2 = XSTEP
        IF(YSTEP <= 0.08)then
          YSTEP2 = YSTEP
          YSTEP = YSTEP + YSTEP
        END IF
        DO
          XMAX(K) = XMAX(K) + XSTEP2
          IF(XMAX(K) < RANGE(1, K) .OR. &
             XMAX(K) > RANGE(2, K)) EXIT
          FMAX = FUNC(XMAX, 2, NINT(XMAX(1)))
          IF(FMAX + 0.01_DP < CI95(0, 2, K)) GOTO 1
          X1 = ABS(FMAX - FMAX2)
          IF(X1 > YSTEP + YSTEP)THEN
            IF(ABS(XSTEP2) > XIII(K, K) + 1.E-5) THEN
              XMAX(K) = XMAX(K) - XSTEP2
              XSTEP2 = REAL(MAX(ABS(XSTEP2 * .5_DP), XIII(K, K)))
              XSTEP2 = SIGN(XSTEP2, XSTEP)
            ELSE
              EXIT
            END IF
          ELSEIF(X1 > YSTEP)THEN
            XSTEP2 = XSTEP2 * .5
            EXIT
          ELSE IF(X1 < YSTEP2)THEN
            XTEM = REAL(MIN(XMAX(K) * .1, DBLE(ABS(XSTEP2 + XSTEP2))))
            XTEM = REAL(MAX(DBLE(XTEM), XIII(K, K)))
            XSTEP2 = SIGN(XTEM, XSTEP2)
          END IF
        END DO
        XSTEP = XSTEP2
        IF((XMAX(K) < RANGE(1, K) .AND. &
            CI95(0, 1, k) < 1.1 * RANGE(1, K)) .OR.  &
           (XMAX(K) > RANGE(2, K) .AND. &
            CI95(0, 1, k) > .99 * RANGE(2, K))) EXIT
        IF(XMAX(K) < RANGE(1, K))THEN
          XMAX(K) = RANGE(1, K)
          KFLAG = 1
        ELSEIF(XMAX(K) > RANGE(2, K))THEN
          XMAX(K) = RANGE(2, K)
          KFLAG = 1
        ELSE
          KFLAG = -1
        END IF
        IF(M_ESTIMATE == 1)THEN
          XI = XIII
          CALL POWELL(XMAX, XI, ITER, FMAX, &
                      M_ESTIMATE, K, MONITOR)
          IF(FMAX + 0.01_DP < CI95(0, 2, K)) GOTO 1
        ELSE
          IF(KFLAG == 1)FMAX = FUNC(XMAX, 2, NINT(XMAX(1)))
        END IF
        KK = KK + KSTEP
        CI95(KK, 2, K) = FMAX
        CI95(KK, 1, K) = XMAX(K)
        X1 = FMAX - 2._DP - CI95(0, 2, K)
        IF(X1 > -0.001_DP .AND. X1 < 0.001_DP)THEN
          BOUND(K2, K) = KK
          EXIT
        ELSEIF(X1 >= 0.001_DP .OR. X1 > -0.1)THEN
          IF(ABS(KK) > 2)THEN
            NPL = 3
          ELSE
            NPL = 2
          END IF
          M1 = KK - (NPL - 1) * KSTEP
          XA(1 : NPL) = CI95(KK : M1 : -KSTEP, 2, K)
          YA(1 : NPL) = CI95(KK : M1 : -KSTEP, 1, K)
          X1 = CI95(0, 2, K) + 2._DP
          CALL polint(xa, ya, NPL, x1, y1, dy)
          CI95(KK, 1, K) = Y1
          CI95(KK, 2, K) = X1
          BOUND(K2, K) = KK
          EXIT
        END IF
        IF(KFLAG == 1)EXIT
        XMAX2 = XMAX
        FMAX2 = FMAX
      END DO
    END DO
  END DO
!Output analysis results
  IF(Thd0Q) CALL Output(BOUND, CI95, RANGE, MTNE, MTM)
  IF(NumMPI > 1) THEN
!$  CALL MPI_INITIALIZED(ExistQ, K)
!$  IF(ExistQ) CALL MPI_FINALIZE(K)
  END IF
END PROGRAM MAIN
!
Subroutine GetRunCmmd
  Use Module_Data
  IMPLICIT NONE
  INTEGER :: I, J, K, M, N
  CHARACTER(Len = 999) :: C999
  CHARACTER(Len = 5) :: HELP(2, 2, 2, 2)
  CHARACTER(Len = 1) :: H(2, 4)
  DATA H/'h', 'H', 'e', 'E', 'l', 'L', 'p', 'P'/
  FORALL(I = 1 : 2, J = 1 : 2, K = 1 : 2, M = 1 : 2) &
    HELP(I, J, K, M) = '-' // H(I, 1) // H(J, 2) // &
    H(K, 3) // H(M, 4)
  ParaFile = 'MNE_DATA'      !Default
  MPIRUN = 0                 !Default
  OutFreqEstQ = 1            !Default
  DO I = 1, 3
    CALL Get_command_argument(I, C999)
    IF(LEN_TRIM(C999) == 0) EXIT
    IF(ANY(TRIM(C999) == HELP(:, :, :, :))) THEN
      CALL ShowVersion(6)
      WRITE(*,'(A)') 'Command Line Arguments:', &
        '  Str:[Input parameter file name]', &
        '  1/0:[Using MPI = Y/N]', &
        '  1/0:[Write allele frequency to file = Y/N]'
      WRITE(*,'(/A)') 'Details of the software ' // &
        'in user manual, MLNeUserGuide.pdf'
      STOP
    END IF
    IF(I == 1) THEN
      ParaFile = TRIM(AdjustL(C999))
    ELSE IF(I == 2) THEN
      READ(C999, '(I12)', IOSTAT = N) M
      IF(N /= 0) THEN
        WRITE(*, *)'Error reading command for using ' // &
          'MPI (1/0 = Y/N)!'
        STOP
      ELSE
        MPIRUN = Merge(0, 1, M == 0)
      END IF
    ELSE
      READ(C999, '(I12)', IOSTAT = N) M
      IF(N /= 0) THEN
        WRITE(*, *)'Error reading command for output ' // &
          'allele frequency estimates (1/0 = Y/N)!'
        STOP
      ELSE
        OutFreqEstQ = Merge(0, 1, M == 0)
      END IF
    END IF
  END DO
END Subroutine GetRunCmmd
!
Subroutine Output(BOUND, CI95, RANGE, MTNE, MTM)
  USE Module_Data
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: BOUND(2, 2)
  INTEGER :: I, J, K, L, M, N, NN
  REAL(DP), INTENT(IN) :: CI95(-100 : 100, 2, 2), RANGE(2, 2)
  REAL(SP), INTENT(IN) :: MTNE, MTM
  CHARACTER(Len = 15) :: C15(2)
  CHARACTER(Len = :), Allocatable :: Str, Line
  CHARACTER(Len = 7), Allocatable :: RC(:)
  ALLOCATE(Character(Len = 1) :: Str)
  CALL ShowVersion(10)
  OPEN(UNIT = 10, FILE = OutFile, POSITION = 'APPEND')
  WRITE(10, '(T19, A/)')'Output from the program'
  Str = '--------------------------------'
  WRITE(10, '(/A/A/A)') Str, &
    'Parameters effective in analysis', Str
  IF(M_ESTIMATE == 0)THEN
    Str = 'Estimating Ne only, ' // &
      'a single isolated population'
  ELSE
    Str = 'Estimating Ne and m jointly, ' // &
      'assuming an infinite source'
  END IF
  WRITE(10, '(A)') 'Model : ' // Str
  IF(M_ESTIMATE == 1) WRITE(10, '(A)') &
    'Equilibrium : ' // ItoC(I_Equilibrium)
  WRITE(10, '(A)') 'MaximumNe : ' // ItoC(MaximumNe), &
    'Monitor : ' // ItoC(Monitor), &
    'openMP Threads/(MPI process) : ' // ItoC(NumThread), &
    'MPI Processes : ' // ItoC(NumMPI), &
    '#Loci : ' // ItoC(Loci_TT), &
    'DataFile : ' // GtypeFile, &
    'ParaFile : ' // ParaFile, &
    'DataForm : ' // ItoC(DataForm), &
    '#Temporal Samples : ' // ItoC(NSAMB)
  WRITE(10, '(A)', Advance = 'NO' ) '#Generations :'
  WRITE(10, '(999A)') (' ' // ItoC(NGEN(I)), I = 1, NSAMB)
  WRITE(10, '(A)', Advance = 'NO' ) 'Sample Size(Focal) :'
  WRITE(10, '(999A)') (' ' // ItoC(NumIndiv(I)), I = 1, NSAMB)
  WRITE(10, '(/A)') 'Genotype data'
  WRITE(10, '(A)') '  Number polymorphic loci = ' // &
    ItoC(NPoly), '  Number monomorphic loci = ' // &
    ItoC(NMono)
  IF(M_ESTIMATE == 1) WRITE(10, '(A)') &
    'Sample Size(Source) : ' // ItoC(NumIndiv(0)), &
    'INIT_POINT : ' // ItoC(INIT_POINT)
  Str = '****************'
  WRITE(10, '(/A/A9,3A15)') Str // &
    ' Likelihood estimates ' // Str, &
    'PARAMETER', 'POINT_EST', '+95%CI', '-95%CI'
  WRITE(10, '(A9)', Advance = 'NO') 'Ne'
  IF(ALL(BOUND(:, 1) /= 0)) THEN
    WRITE(10, '(3F15.2)') CI95(0, 1, 1) / 2, &
      CI95(BOUND(1, 1), 1, 1) / 2, &
      CI95(BOUND(2, 1), 1, 1) / 2
  ELSEIF(BOUND(1, 1) /= 0) THEN
    C15(1) = '<' // RtoC(RANGE(1, 1) / 2, '(F12.2)')
    WRITE(10, '(2F15.2, A)') CI95(0, 1, 1) / 2, &
      CI95(BOUND(1, 1), 1, 1) / 2, AdjustR(C15(1))
  ELSEIF(BOUND(2, 1) /= 0) THEN
    C15(1) = '>' // RtoC(RANGE(2, 1) / 2, '(F12.2)')
    WRITE(10, '(F15.2, A, F15.2)') &
      CI95(0, 1, 1) / 2, AdjustR(C15(1)), &
      CI95(BOUND(2, 1), 1, 1) / 2
  ELSE
    C15(1) = '>' // RtoC(RANGE(2, 1) / 2, '(F12.2)')
    C15(2) = '<' // RtoC(RANGE(1, 1) / 2, '(F12.2)')
    WRITE(10, '(F15.2, 2A)') CI95(0, 1, 1), &
      AdjustR(C15(1 : 2))
  END IF
  IF(M_ESTIMATE == 1) THEN
    WRITE(10, '(A9)', Advance = 'NO') 'm'
    IF(ALL(BOUND(:, 2) /= 0))THEN
      WRITE(10, '(3F15.4)') CI95(0, 1, 2), &
        CI95(BOUND(1, 2), 1, 2), CI95(BOUND(2, 2), 1, 2)
    ELSEIF(BOUND(1, 2) /= 0)THEN
      C15(1) = '<' // RtoC(RANGE(1, 2), '(F12.4)')
      WRITE(10, '(2F15.4, A)') CI95(0, 1, 2), &
        CI95(BOUND(1, 2), 1, 2), AdjustR(C15(1))
    ELSEIF(BOUND(2, 2) /= 0)THEN
      C15(1) = '>' // RtoC(RANGE(2, 2), '(F12.4)')
      WRITE(10, '(F15.4, A, F15.4)') CI95(0, 1, 2), &
        AdjustR(C15(1)), CI95(BOUND(2, 2), 1, 2)
    ELSE
      C15(1) = '>' // RtoC(RANGE(2, 2), '(F12.4)')
      C15(2) = '<' // RtoC(RANGE(1, 2), '(F12.4)')
      WRITE(10, '(F15.4, 2A)') CI95(0, 1, 2), &
        AdjustR(C15(1:2))
    END IF
  END IF
  WRITE(10, '(/A/2A16)') 'Relative profile ' // &
    'log-likelihood as a function of Ne', 'Ne', 'Log(L)'
  DO I = -100, 100
    IF(ABS(CI95(I, 1, 1)) > 1.E-50_DP) &
      WRITE(10, '(F16.2, F16.4)') CI95(I, 1, 1) / 2, &
        CI95(0, 2, 1) - CI95(I, 2, 1)
  END DO
  IF(M_ESTIMATE == 1)THEN
    WRITE(10, '(/A/2A16)') 'Relative profile ' // &
      'log-likelihood as a function of  m', 'm', 'Log(L)'
    DO I = -100, 100
      IF(ABS(CI95(I, 1, 2)) > 1.E-50_DP) &
        WRITE(10, '(2F16.4)') CI95(I, 1, 2), &
          CI95(0, 2, 2) - CI95(I, 2, 2)
    END DO
  END IF
  WRITE(10, '(/A/)')Str // ' Moment estimates ' // Str
  WRITE(10, '(A9,A12)') 'PARAMETER', 'POINT_EST'
  WRITE(10, '(A9, F12.2)') AdjustR('Ne'), MTNE
  IF(M_ESTIMATE == 1)THEN
    WRITE(10, '(A9, F12.4)') AdjustR('m'), MTM
  ELSE
    WRITE(10, *)
  END IF
  I = 0
  IF(MTNE < 0.) THEN
    WRITE(10, '(A)') NewLn // '  Ne < 0 ' // &
      'indicates an infinitely large estimate of Ne'
    I = 1
  END IF
  IF(M_ESTIMATE == 1 .AND. MTM < 0.) THEN
    IF(I == 0) WRITE(10, *)
    WRITE(10, '(A)')'  m < 0 indicates NO valid m estimate'
  END IF
  IF(OutFreqEstQ == 1) THEN
    WRITE(10, '(//)')
    Str = '**************************'
    WRITE(10, '(A)') Str, 'Allele frequency estimates', Str
    WRITE(10, '(2A)') '#Polymorphic loci = ' // ItoC(NLOCI), &
      '  #Monomorphic loci = ' // ItoC(Loci_TT - NLOCI)
    IF(M_ESTIMATE == 1) THEN
      WRITE(10, '(A)', Advance = 'NO') &
        'Col 1, 2, 3, 3+i are locus ' // &
        'index, allele, freq in source sample and '
    ELSE
      WRITE(10, '(A)', Advance = 'NO') &
        'Col 1, 2, 2+i are locus ' // &
        'index, allele, freq '
    END IF
    WRITE(10, '(A)') 'in focal sample i ' // &
      '(i=1-' // ItoC(NSAMB) // ')'
    N = 0
    Allocate(RC(0 : 10000))
    DO I = 0, 10000
      RC(I) = ' ' // RtoC(I * 0.0001D0, '(F6.4)')
    END DO
    ALLOCATE(Character(Len = 99999) :: Line)
    NN = 0
    DO I = 1, NLOCI
      Str = ItoC(I) // ' '
      M = Len_Trim(Str) + 1
      DO J = 1, KALLELE(I)
        IF(NN > 0) THEN
          K = Len(NewLn)
          Line(NN + 1 : NN + K) = NewLn
          NN = NN + K
        END IF        
        Line(NN + 1 : NN + M) = Str
        NN = NN + M
        C15(1) = ItoC(AllAllele(N + J))
        L = Len_Trim(C15(1))
        Line(NN + 1 : NN + L) = Trim(C15(1))
        NN = NN + L
        DO K = 1 - M_ESTIMATE, NSAMB
          Line(NN + 1 : NN + 7) = &
            RC(NINT((10000.0 * NSAMP_T(N + J, K)) / &
            Sum(NSAMP_T(N + 1 : N + KALLELE(I), K))))
          NN = NN + 7
        END DO
      END DO
      IF(NN > 99799) THEN
        WRITE(10, '(A)') Line(1 : NN)
        NN = 0
      END IF
      N = N + KALLELE(I)
    END DO
    IF(NN > 0) WRITE(10, '(A)') Line(1 : NN)
  END IF
  WRITE(10, '(//)')
  DO I = 1, 2
    IF(I == 2) CALL DATE_AND_TIME(DATE, TIME)
    WRITE(10, '(A)') 'Run ' // Merge('started ', &
      'finished', I == 1) // ' date = ' // &
      Day_of_week(Date) // ' ' // &
      Date(7:8) // '/' // Date(5:6) // '/' // &
      Date(1:4) // ', Time = ' // Time(5:6) // &
      '/' // Time(3:4) // '/' // Time(1:2)
  END DO
  CLOSE(10)
  WRITE(*, '(/A)')'Run finished!' // NewLn // &
    'Output to file: ' // OutFile
END Subroutine Output
!
Subroutine InitializeThreads
  USE Module_Data
  IMPLICIT NONE
!$  Include 'mpif.h'
!$  INTEGER :: I
  ThreadID = 0
  NumMPI = 1
  IF(MPIRUN == 1) THEN
!$  CALL MPI_INIT(I)
!$  IF(I /= MPI_SUCCESS)THEN
!$    WRITE(*,'(A)')'Error starting MPI program. ' // &
!$      'Terminating.'
!$    STOP
!$  ELSE
!$    CALL MPI_COMM_RANK(MPI_COMM_WORLD, ThreadID, I)
!$    CALL MPI_COMM_SIZE(MPI_COMM_WORLD, NumMPI, I)
!$  END IF
  END IF
  Thd0Q = (ThreadID == 0)
END Subroutine InitializeThreads
!
Subroutine Sleeping(snd)
  implicit none ! desired sleep time in second
  integer :: snd, t1, t2, WallTime
  t1 = WallTime(0)
  do
    t2 = WallTime(0)
    if(t2 - t1 >= snd) exit
  enddo
End Subroutine Sleeping
!
Function WallTime(M) !Wall time in seconds or milliseconds
  implicit none
  INTEGER, INTENT(IN) :: M
  integer :: t(8), day, WallTime, hour
  Data day/0/, hour/0/
  call date_and_time(values = t)
  if(t(5) < hour) day = day + 1
  hour = t(5)
  WallTime = (day * 24 + t(5)) * 3600 + t(6) * 60 + t(7)
  IF(M == 1) WallTime = WallTime * 1000 + T(8)
End Function WallTime
!
Subroutine ShowVersion(Unit)
  USE Module_Data
  IMPLICIT NONE
  INTEGER(I4B) :: I, Unit
  CHARACTER(Len = 82) :: Blnk82
  Character(Len = 65) :: PrntLine
  FORALL(I = 1 : 82) Blnk82(I : I) = ' '
  FORALL(I = 1 : 65) PrntLine(I : I) = '-'
  IF(Unit /= 6) OPEN(Unit, File = OutFile)
  WRITE(Unit, '(A/A)') PrntLine, Blnk82(1:13) // &
    'MLNe, Version ' // Version // ', Build ' // Build
  WRITE(Unit, '(A)') Blnk82(1:14) // 'Copyright (C) 2021-' // &
    Build(1 : 4) // ' Jinliang Wang', Blnk82(1:8) // &
    'Institute of Zoology, Zoological Society of London'
  WRITE(Unit, '(A/A/)') Blnk82(1:18) // &
    'Email: jinliang.wang@ioz.ac.uk', PrntLine
  IF(Unit /= 6) CLOSE(Unit)
END Subroutine ShowVersion
!
FUNCTION FstFunc(X)
  Use Module_Data
  IMPLICIT NONE
  REAL(DP), intent(in) :: X
  INTEGER, Parameter :: NPoint = 50
  REAL(DP), Parameter :: RPoint = 1._DP / NPoint, TOL = 1.E-5_DP
  REAL(DP) :: FstFunc, F(0 : 2), PI, P, X0, XI, X1(0 : 1), &
    X2, XLog, XMax(2), XIMax(2), GAMMLN
  REAL(SP) :: S
  INTEGER :: I, L, NA1(0 : 1), NA2(0 : 1), IA(0 : 1), &
    I1, I2, Idx(3, 2, 0 : 1), J1, J2
  REAL(DP), Save :: LogX(0 : NPoint)
  external :: GAMMLN
  DATA LogX/1.D0, NPoint * 1.D0/
  IF(LogX(0) > 0.D0) THEN
    LogX(0) = Log(0.1 * RPoint)
    LogX(NPoint) = Log(1.0 - 0.1 * RPoint)
    DO I = 1, NPoint - 1
      LogX(I) = Log(I * RPoint)
    END DO
  END IF
  IF(Thd0Q) WRITE(*, '(A,ES11.4,$)') '  4Nm=', X
  IF(X < 0.01_DP .OR. X > 1000._DP) THEN
    FstFunc = 1.E100_DP
    IF(Thd0Q) WRITE(*, '(A,ES11.4)') '  LogL=', FstFunc
    RETURN
  END IF
  X0 = GAMMLN(X)
  F(0) = 0._DP
  DO L = 1, NTLOCI
    DO I = 0, 1
      NA1(I) = NSamp(L, I)
      NA2(I) = NS2(L, I) - NA1(I)
      IA(I) = NINT(NA1(I) * NPoint / DBLE(NS2(L, I)))
      IF(IA(I) <= 0) THEN
        IA(I) = 1
      ELSE IF(IA(I) >= NPoint) THEN
        IA(I) = NPoint - 1
      END IF
      Idx(:, 1, I) = (/IA(I), 0, -1/)
      Idx(:, 2, I) = (/IA(I) + 1, NPoint, 1/)
      X1(I) = FAC(NS2(L, I)) - FAC(NA1(I)) - FAC(NA2(I)) - &
           NS2(L, I) * (FAC(NPoint) - FAC(NPoint - 1))
    END DO
    XIMax(:) = -1._DP
    F(1) = 0._DP
    DO I1 = 1, 2
      DO I2 = Idx(1, I1, 0), Idx(2, I1, 0), Idx(3, I1, 0)
        IF(I2 == 0) THEN
          XI = 0.1_DP * RPoint
        ELSE IF(I2 == NPoint) THEN
          XI = 1._DP - 0.1_DP * RPoint
        ELSE
          XI = I2 * RPoint
        END IF
        X2 = X0 - GAMMLN(X * XI) - GAMMLN(X * (1._DP - XI))
        CALL BINOM(PI, XLog, X1(0), I2, NPoint, NA1(0), NA2(0))
        F(2) = 0._DP
        XMax(:) = -1._DP
        DO J1 = 1, 2
          DO J2 = Idx(1, J1, 1), Idx(2, J1, 1), Idx(3, J1, 1)
            CALL BINOM(P, XLog, X1(1), J2, NPoint, NA1(1), NA2(1))
            P = X2 + (X * XI - 1._DP) * LogX(J2) + &
                     (X * (1._DP - XI) - 1._DP) * LogX(NPoint - J2)
            P = Exp(P + XLog)
            F(2) = F(2) + P
            IF(P < XMax(2)) THEN
              EXIT
            ELSE IF(P < XMax(1)) THEN
              XMax(1) = P
              XMax(2) = P * TOL
            END IF
          END DO
        END DO
        P = F(2) * PI
        F(1) = F(1) + P
        IF(P < XIMax(2)) THEN
          EXIT
        ELSE IF(P < XIMax(1)) THEN
          XIMax(1) = P
          XIMax(2) = P * TOL
        END IF
      END DO
    END DO
    F(0) = F(0) + Log(F(1)) * FACTOR(L)
  END DO
  FstFunc = -F(0)
  IF(Thd0Q) WRITE(*, '(A,ES11.4)') '  LogL=', FstFunc
END FUNCTION FstFunc
!
SUBROUTINE MTEST(MTNE, MTM)
  Use Module_Data
  REAL(SP), PARAMETER :: TINY = 1.E-8, BIG = 1. - TINY
  REAL(SP) :: MTNE, MTM, MTNE2, MTM2
  MTNE = 0.
  MTM = 0.
  NN = 0
  DO KK = 1, NSAMB - 1
    KK1 = KK + 1
    IF(M_ESTIMATE == 1)THEN
      DBT = 0.
      WNE = 0.
      DAB0 = 0.
      WW = 0.
      WF = 0.
    ELSE
      FC = 0.
      NLOCI_FC = 0
    END IF
    DO I = 1, NTLOCI
      B0 = NSAMP(I, KK) / REAL(NS2(I, KK))
      BT = NSAMP(I, KK1) / REAL(NS2(I, KK1))
      IF(M_ESTIMATE == 1)THEN
        A0 = NSAMP(I, 0) / REAL(NS2(I, 0))
        XT = abs(A0 - B0)
        DAB0 = DAB0 - XT * XT
        IF(XT > TINY)THEN
          WI = (A0 - B0) ** 2
          WW = WW + WI
          FI = (BT - B0) / (A0 - B0)
          WF = WF + WI * FI
        END IF
        PQ = (B0 * (1. - B0) + BT * (1. - BT)) * 0.5
        WNE = WNE + PQ
        DBT = DBT + (BT - B0) ** 2 - B0 * (1 - B0) / &
          NS2(I, KK) - BT * (1. - BT) / NS2(I, KK1)
      ELSE
        XT = BT + B0
        IF(XT >= TINY .AND. XT < 2.)THEN
          FC = FC + 2. * (BT - B0) ** 2 * (1. / XT + &
            1. / (2. - XT)) - 1. / NS2(I, KK) - 1. / NS2(I, KK1)
          NLOCI_FC = NLOCI_FC + 1
        END IF
      END IF
    END DO
    NT = NGEN(KK1) - NGEN(KK)
    IF(M_ESTIMATE == 1)THEN
      IF(WW <= 0.)CYCLE
      NN = NN + 1
      MTM2 = 1. - ABS(1. - WF / WW) ** (1. / nt)
      MTM = MTM + MTM2
      IF(ABS(MTM2) > TINY)THEN
        MTNE2 = (DBT + DAB0 * (1. - &
          (1. - MTM2) ** NT) ** 2) / (WNE * &
          (1. - (1. - MTM2) ** (2 * NT)) / &
          (MTM2 * (2. - MTM2)))
      ELSE
        MTNE2 = DBT / (WNE * NT)
      END IF
    ELSE
      MTNE2 = FC / NLOCI_FC / REAL(NT)
      NN = NN + 1
    END IF
    IF(MTNE2 > 0.)MTNE = MTNE + MTNE2
  END DO
  IF(NN == 0)THEN
    MTNE = -1.
    MTM = -10.
  ELSE
    IF(MTNE > TINY)THEN
      MTNE = NN / MTNE * .5
    ELSE
      MTNE = -1.
    END IF
    IF(M_ESTIMATE == 1) MTM = MTM / NN
  END IF
END SUBROUTINE MTEST
!
SUBROUTINE GTDIS(NINTPA_FLAG)
  Use Module_Data
  REAL(DP) :: X2, X3, xmax, xmax2, YLOG
  INTEGER :: LSTEP(2), LBOUND(2), &
    NP(NSAMB), NPT(NSAMB), NP0, NP1
  DATA LSTEP/-1, 1/, LBOUND/0, 1/
  LBOUND(2) = NINTPA
  DO L = 1, NTLOCI
    NSPA = NS2(L, 0)
    N1 = NSAMP(L, 0)
    N2 = NSPA - N1
    X2 = FAC(NSPA) - NSPA * (FAC(NINTPA) - &
      FAC(NINTPA - 1)) - FAC(N1) - FAC(N2)
    XMAX = -1._DP
    NI = NINT(N1 * NINTPA / DBLE(NSPA))
    IF(NI <= 0)THEN
      NI = 1
    ELSEIF(NI >= NINTPA)THEN
      NI = NINTPA - 1
    END IF
    FORALL(K = 1 : NSAMB)
      NP(K) = NSAMP(L, K)
      NPT(K) = NS2(L, K)
    END FORALL
    NP0 = Merge(1, 0, ANY(NP(1 : NSAMB) > 0))
    NP1 = Merge(1, 0, ANY(NP(1 : NSAMB) < NPT(1 : NSAMB)))
    DO JJ = 1, 2
      IBPPAA(L, JJ) = LBOUND(JJ)
      DO K = NI + JJ - 1, LBOUND(JJ), LSTEP(JJ)
        CALL BINOM(X3, YLOG, X2, K, NINTPA, N1, N2)
        IF(K == 0 .AND. N1 == 0)THEN
          IF(NP0 == 1)THEN
            X3 = 0._DP
            YLOG = -1.E300_DP
          END IF
        ELSE IF(K == NINTPA .AND. N1 == NSPA)THEN
          IF(NP1 == 1)THEN
            X3 = 0._DP
            YLOG = -1.E300_DP
          END IF
        END IF
        PPAA(K, L) = X3
        IF(X3 > XMAX)THEN
          XMAX = X3
          XMAX2 = YLOG + PREC_LOG
          JMAX = K
        ELSEIF(YLOG < XMAX2)THEN
          IBPPAA(L, JJ) = K
          EXIT
        END IF
      END DO
    END DO
    XMAX2 = XMAX * PREC
    DO JJ = JMAX, LBOUND(1), -1
      IF(PPAA(JJ, L) < XMAX2)THEN
        IBPPAA(L, 1) = JJ
        EXIT
      END IF
    END DO
  END DO
  NINTPA_FLAG = 0
  DO L = 1, NTLOCI
    X3 = SUM(PPAA(IBPPAA(L, 1) : IBPPAA(L, 2), L))
    IF(X3 > 1.E-30_DP) THEN
      X3 = 1._DP / X3
      DO i = IBPPAA(L, 1), IBPPAA(L, 2)
        PPAA(I, L) = PPAA(I, L) * X3
      END DO
    ELSE
      NINTPA_FLAG = 1
      EXIT
    END IF
  END DO
END SUBROUTINE GTDIS
!
FUNCTION func(XMIN, NP, IDIM)
  Use Module_Data
!$  Include 'mpif.h'
  REAL(DP) :: X0, X1, X2, X3, XT, Y, XMAX, &
    XMAX2, XMIN(NP), XDIM, FUNC, YLOG, GAMMLN
  REAL(DP), SAVE :: Scaler
  REAL(DP), Allocatable :: FUNCL(:), PBDIS(:), PT(:, :)
  INTEGER :: IA(2), LSTEP(2), LBOUNDS(2), &
    MAXDIM, Ndx, ID, NN, N1
  INTEGER, PARAMETER :: CHUNK = 1
  INTEGER, Allocatable :: JA(:), IFLAG(:), Idx(:)
  LOGICAL :: MonitorQ
  INTEGER :: MyTime, WallTime
!$  INTEGER, EXTERNAL :: omp_get_thread_num
  SAVE XDIM, MAXDIM
  DATA LSTEP/-1, 1/, LBOUNDS/0, 1/, Scaler/-1.D0/
  EXTERNAL :: GAMMLN
  IF(Scaler < 0.D0) Scaler = Merge(1.D0, &
    Dble(NPoly) ** 0.35 - 7.8, NPoly <= 500)
  Allocate(IFLAG(MAXLC), Idx(NTLOCI), &
    FUNCL(MAXLC), PBDIS(0 : IDIM))
  MonitorQ = (MONITOR > 1 .AND. IDIM > 5000 .OR. MONITOR > 2)
  IF(MonitorQ) MyTime = WallTime(0)
  IF(M_ESTIMATE == 0)XMIN(2) = 0._DP
  IF(XMIN(1) < 2 .OR. XMIN(1) > NE20 .OR. XMIN(2) < 0._DP &
       .OR. XMIN(2) > 1._DP)THEN
    FUNC = 1.E100_DP
    RETURN
  END IF
  N1 = NINT(XMIN(1) * (REAL(NGEN(NSAMB))) ** 0.75)
  IMM = NINT(XMIN(2) * 1000)
  DO I = 1, NFUNC
    IF(IDIM == NPARA(I, 1) .AND. IMM == NPARA(I, 2))THEN
      FUNC = FPARA(I)
      RETURN
    END IF
  END DO
  LBOUNDS(2) = IDIM
  IF(NPARA(NFUNC, 1) /= IDIM)THEN
    XDIM = 1._DP / IDIM
    DO I = 1, NSAMB
      DO J = 1, NTLOCI
        NNN = NS2(J, I)
        YFC_TEM(J, I) = FAC(NNN) - &
          (FAC(IDIM) - FAC(IDIM - 1)) * NNN
      END DO
    END DO
    CALL TrmxDim(IDIM, XDIM, XMIN(2), MAXDIM)
  END IF
  ALLOCATE(TRMX(0 : MAXDIM), IB(3, 0 : IDIM), STAT = KK)
  IF(KK /= 0)THEN
    IF(Thd0Q) WRITE(*, '(a/a/A)') &
      'The maximum value of Ne ' // &
      'set in your data is too large', &
      '  that your computer has not enough RAM to ' // & 
      'run the program. ', '  Try a smaller maximum ' // &
      'value of Ne and re-run the program'
    STOP
  END IF
  ALLOCATE(PT(0 : IDIM, 0 : 1))
  FUNCL(:) = 0._DP
  Allocate(JA(0 : NumMPI))
  JA(0) = 0
  IF(M_ESTIMATE == 0 .OR. I_Equilibrium == 0)THEN
    PBDIS = 1._DP / (IDIM + 1)
  ELSE
    X0 = Est_4Nm
    X3 = GAMMLN(X0)
  END IF
  DO IPA = 0, NINTPA
    IF(M_ESTIMATE == 1 .AND. I_Equilibrium /= 0)THEN
!             When at drift-migration equilibrium
      IF(IPA > 0 .AND. IPA < NINTPA)THEN
        XT = DBLE(IPA) / NINTPA
        X1 = X0 * XT - 1._DP
        X2 = X0 * (1._DP - XT) - 1._DP
        PBDIS(0) = EXP(X3 - GAMMLN(X1 + 2.D0) - &
          GAMMLN(X2 + 1.D0)) * XDIM ** (X1 + 1.D0)
        PBDIS(IDIM) = EXP(X3 - GAMMLN(X1 + 1.D0) - &
          GAMMLN(X2 + 2.D0)) * XDIM ** (X2 + 1.D0)
        Y = PBDIS(0) + PBDIS(IDIM)
        XT = EXP(X3 - GAMMLN(X1 + 1.D0) - &
          GAMMLN(X2 + 1.D0)) * XDIM
        DO I = 1, IDIM - 1
          PBDIS(I) = XT * (I * XDIM) ** X1 * &
                   (1.D0 - I * XDIM) ** X2
          Y = Y + PBDIS(I)
        END DO
        Y = 1._DP / Y
        DO I = 0, IDIM
          PBDIS(I) = PBDIS(I) * Y
        END DO
      ELSEIF(IPA == 0)THEN
        PBDIS(0) = 1.D0
        PBDIS(1 : ) = 0.D0
      ELSE
        PBDIS(IDIM) = 1.D0
        PBDIS(0 : IDIM - 1) = 0.D0
      END IF
    END IF
    IF(XMIN(2) > 0._DP)THEN
      Ndx = 0
      DO I = 1, NTLOCI
        IF(IPA >= IBPPAA(I, 1) .AND. &
           IPA <= IBPPAA(I, 2)) THEN
          IFLAG(I) = 1
          Ndx = Ndx + 1
          Idx(Ndx) = I
        ELSE
          IFLAG(I) = 0
        END IF
      END DO
      IF(Ndx == 0)CYCLE
    ELSE
      Ndx = NTLOCI
      FORALL(I = 1 : NTLOCI) Idx(I) = I
    END IF
    IF(Thd0Q .AND. MonitorQ) WRITE(*, '(A,$)') &
      Char(13) // 'Ne = ' // ItoC(IDIM / 2) // &
      ' Trans-Matrix...'
    CALL GTRANS(IDIM, XDIM, XMIN(2), IPA, MAXDIM)
    DO I = 1, NumMPI
      JA(I) = Ndx / NumMPI
      IF(I <= Mod(Ndx, NumMPI)) JA(I) = JA(I) + 1
    END DO
    DO I = 1, NumMPI
      JA(I) = JA(I) + JA(I - 1)
    END DO
! Open MP parallel for multiple processors
    ID = 0
!$OMP PARALLEL IF(NumThread > 1) NUM_THREADS(NumThread) &
!$OMP   DEFAULT(SHARED) Private(II, I, NA1, NA2, JM, XMAX, X1, &
!$OMP   X2, K, IA, J, YLOG, JMAX, XMAX2, KFLAG, KK, &
!$OMP   XT, MARK, NG, PT, NN) FirstPrivate(ID)
!$  ID = omp_get_thread_num()
    NN = 0
!$OMP DO SCHEDULE(dynamic, CHUNK)
    DO II = JA(ThreadID) + 1, JA(ThreadID + 1)
      I = Idx(II)
      IF(Thd0Q .AND. MonitorQ .AND. ID == 0) THEN
        NN = NN + N1
        IF(NN > 10000000) THEN
          NN = 0
          IF(M_ESTIMATE == 0) THEN
            WRITE(*, '(A,$)') Char(13) // 'Ne = ' // &
              ItoC(IDIM / 2) // '      Locus = ' // ItoC(I)
          ELSE
            WRITE(*, '(A,$)') Char(13) // 'Ne = ' // &
              ItoC(IDIM / 2) //  '  m = ' // &
              RtoC(XMIN(2), '(F8.4)') // &
              '  Locus = ' // ItoC(I)
          END IF
        END IF
      END IF
      NA1 = NSAMP(I, 1)
      NA2 = NS2(I, 1) - NA1
      JM = NINT(NA1 * IDIM / DBLE(NS2(I, 1)))
      IF(JM <= 0)THEN
        JM = 1
      ELSEIF(JM >= IDIM)THEN
        JM = IDIM - 1
      END IF
      XMAX = 1.D-200
      X2 = YFC_TEM(I, 1) - FAC(NA1) - FAC(NA2)
      DO K = 1, 2
        IA(K) = LBOUNDS(K)
        DO J = JM + K - 1, LBOUNDS(K), LSTEP(K)
          CALL BINOM(X1, YLOG, X2, J, IDIM, NA1, NA2)
          X1 = X1 * PBDIS(J)
          PT(J, 0) = X1
          IF(X1 > XMAX)THEN
            XMAX = X1
            JMAX = J
            XMAX2 = X1 * PREC
          ELSEIF(X1 < XMAX2)THEN
            IF(X1 < 1.D-200) THEN
              IA(K) = J - LSTEP(K)
            ELSE
              IA(K) = J
            END IF
            EXIT
          END IF
        END DO
      END DO
      IF(XMAX <= 1.1D-200) CYCLE
      DO NG = 1, NGEN(NSAMB)
        CALL DISTR(IDIM, JMAX, IA, IDIM, &
          PT(0 : IDIM, 0), PT(0 : IDIM, 1))
        KFLAG = 0
        DO KK = 2, NSAMB
          IF(NGEN(KK) /= NG)CYCLE
          KFLAG = 1
          JM = JMAX
          IF(JM <= 0)THEN
            JM = 1
          ELSEIF(JM >= IDIM)THEN
            JM = IDIM - 1
          END IF
          NA1 = Nsamp(I, KK)
          NA2 = NS2(I, KK) - NA1
          XT = YFC_TEM(I, KK) - FAC(NA1) - FAC(NA2)
          XMAX = -1._DP
          DO K = 1, 2
            DO J = JM + K - 1, IA(K), LSTEP(K)
              CALL BINOM(X1, YLOG, XT, J, IDIM, NA1, NA2)
              X2 = PT(J, 1) * X1
              PT(J, 0) = X2
              IF(X2 > XMAX)THEN
                XMAX = X2
                JMAX = J
                XMAX2 = X2 * PREC
              ELSEIF(X2 < XMAX2)THEN
                IA(K) = J
                EXIT
              END IF
            END DO
          END DO
        END DO
        IF(KFLAG == 0) PT(IA(1) : IA(2), 0) = &
                       PT(IA(1) : IA(2), 1)
      END DO
      X2 = SUM(PT(IA(1) : IA(2), 0))
      IF(XMIN(2) > 0._DP)THEN
        FUNCL(I) = FUNCL(I) + X2 * PPAA(IPA, I)
      ELSE
        FUNCL(I) = X2
      END IF
    END DO
!$OMP END DO NOWAIT
!$OMP END PARALLEL
    IF(XMIN(2) <= 0._DP) EXIT
  END DO
  DEALLOCATE(TRMX, PT, IB)
  FUNC = 100.0D0 * Log(1.D7 + XMIN(1)) / NumMPI
  DO II = JA(ThreadID) + 1, JA(ThreadID + 1)
    I = Idx(II)
    IF(FUNCL(I) > 1.E-300_DP)THEN
      FUNC = FUNC - LOG(FUNCL(I)) * FACTOR(I)
    ELSE
      FUNC = 1.E100_DP
      EXIT
    END IF
  END DO
  FUNC = FUNC / Scaler
  IF(NumMPI > 1) THEN
    Y = FUNC
!$  CALL MPI_AllReduce(Y, FUNC, 1, &
!$    MPI_DOUBLE_PRECISION, MPI_SUM, MPI_COMM_WORLD, I)
  END IF
  NFUNC = NFUNC + 1
  NPARA(NFUNC, 1) = IDIM
  NPARA(NFUNC, 2) = IMM
  FPARA(NFUNC) = FUNC
  IF(Thd0Q .AND. MonitorQ)THEN
    MyTime = WallTime(0) - MyTime
    WRITE(*, '(A)', Advance = 'NO') Char(13) // &
      'Ne = ' // ItoC(IDIM / 2)
    IF(M_ESTIMATE == 1) WRITE(*, '(A, F8.4)', &
      Advance = 'NO') '  m = ', XMIN(2)
    WRITE(*, '(A, ES18.10, A)') '  Log(L) = ', -FUNC, &
      '  TimeTaken(S) = ' // ItoC(MyTime)
  END IF
END FUNCTION func
!
SUBROUTINE TrmxDim(IDIM, XDIM, Rmig, MAXDIM)
  Use Module_Data
  REAL(DP) :: X1, XDIM, Rmig
  INTEGER ::  IDIM
  IF(IDIM > 500)THEN
    X1 = PREC_LOG * 2._DP * XDIM + &
      2._DP * (FAC(IDIM / 2) - FAC(IDIM / 2 - 1))
    DO J = IDIM / 2 + 1, IDIM - 1
      IF(FAC(J) - FAC(J - 1) + &
         FAC(IDIM - J) - FAC(IDIM - J - 1) <= X1)THEN
        X1 = (J * 2._DP - IDIM) * IDIM
        EXIT
      END IF
    END DO
    IF(X1 < 2147483646)THEN
      IF(IDIM > 10000) THEN
        MAXDIM = NINT(X1 * (1.D0 / IDIM) ** 0.1)
      ELSE
        MAXDIM = NINT(X1)
      END IF 
      IF(M_ESTIMATE /= 1)THEN
        IF(IDIM < 1000)THEN
          MAXDIM = INT(MAXDIM * 0.89)
        ELSE
          MAXDIM = INT(MAXDIM * 0.8)
        END IF
      ELSE
        IF(Rmig <= 0.1) MAXDIM = INT(MAXDIM * 0.89)
      END IF
    ELSE
      MAXDIM = -1
    END IF
  ELSE
    MAXDIM = (IDIM + 1) * (IDIM + 1)
  END IF
END SUBROUTINE TrmxDim
!
SUBROUTINE DISTR(NR, JMAX, IA, ID, PT0, PT1)
  Use Module_Data
  INTEGER :: NR, JMAX, ID
  REAL(DP), INTENT(IN) :: PT0(0 : ID)
  REAL(DP), INTENT(OUT) :: PT1(0 : ID)
  REAL(DP) :: X1, XMAX, XMAX2
  INTEGER :: IA(2), LSTEP(2), LBOUNDS(2)
  DATA LSTEP/-1, 1/, LBOUNDS/0, 1/
  LBOUNDS(2) = NR
  JJM = JMAX
  k3 = IA(1)
  k4 = IA(2)
  XMAX = -1._DP
  DO KI = 1, 2
    IA(KI) = LBOUNDS(KI)
    DO i = JJM + KI - 1, LBOUNDS(KI), LSTEP(KI)
      L1 = IB(1, I)
      L = IB(2, I) - L1 + 1
      K1 = IB(3, I)
      K2 = K1 + L - 1
      J1 = MAX(K1, k3)
      J2 = MIN(K2, k4)
      KL = L1 - K1
      X1 = 0._DP
!$OMP SIMD REDUCTION(+:X1)
      DO J = J1, J2
        X1 = X1 + TRMX(J + KL) * Pt0(J)
      END DO
      Pt1(i) = X1
      IF(X1 > XMAX)THEN
        XMAX = X1
        JMAX = I
        XMAX2 = X1 * Prec
      ELSEIF(X1 < XMAX2) THEN
        IA(KI) = I
        EXIT
      END IF
    END DO
  END DO
  NN = IA(1)
  DO I = JMAX - 1, NN, -1
    IF(PT1(I) < XMAX2) THEN
      IA(1) = I
      EXIT
    END IF
  END DO
END SUBROUTINE DISTR
!
SUBROUTINE POWELL(P, XI, ITER, FRET, M_ESTIMATE, IVAR, MONITOR)
  Use Module_Data, ONLY : DP, ItoC, RtoC, Thd0Q
  INTEGER, PARAMETER :: N = 2, ITMAX = 200
  REAL(DP) :: DEL, FP, FPTT, T, PT(N), PTT(N), &
    XIT(N), FRET, P(N), XI(N, N), FUNC
  REAL(DP), PARAMETER :: FTOL = 1.E-6_DP
  INTEGER :: MyTime, WallTime
  EXTERNAL FUNC
  FRET = FUNC(P, N, NINT(P(1)))
  PT = P
  ITER = 0
  DO
    ITER = ITER + 1
    FP = FRET
    IBIG = 0
    DEL = 0._DP
    DO I = 1, N
      IF(I /= IVAR)THEN
        IF(MONITOR > 0) MyTime = WallTime(0)
        XIT(1 : N) = XI(1 : N, I)
        FPTT = FRET
        CALL LINMIN(P, XIT, N, FRET, FTOL)
        IF(ABS(FPTT - FRET) > DEL)THEN
          DEL = ABS(FPTT - FRET)
          IBIG = I
        END IF
        IF(Thd0Q .AND. MONITOR > 0) THEN
          MyTime = WallTime(0) - MyTime
          WRITE(*, '(A)') 'Iter = ' // ItoC(ITER) // &
          '  Line minimization = ' // ItoC(I) // &
          '  Log(L) = ' // RtoC(-FRET, '(F16.6)') // &
          '  TimeTaken(S) = ' // ItoC(MyTime)
        END IF
        IF(M_ESTIMATE == 0) RETURN
      END IF
    END DO
    IF(IVAR > 0)RETURN
    IF(2._DP * ABS(FP - FRET) <= &
      FTOL * (ABS(FP) + ABS(FRET))) RETURN
    IF(Thd0Q .AND. ITER == ITMAX) WRITE(*, *) &
      'POWELL EXCEEDING MAXIMUM ITERATIONS'
    PTT(:) = 2._DP * P(:) - PT(:)
    XIT(:) = P(:) - PT(:)
    PT(:) = P(:)
    FPTT = FUNC(PTT, N, NINT(PTT(1)))
    IF(FPTT >= FP)CYCLE
    T = 2._DP * (FP - 2._DP * FRET + FPTT) * (FP - &
        FRET - DEL) ** 2 - DEL * (FP - FPTT) ** 2
    IF(T >= 0._DP)CYCLE
    CALL LINMIN(P, XIT, N, FRET, FTOL)
    IF(IVAR == N)then
      NN = N - 1
    ELSE
      NN = N
    END IF
    XI(:, IBIG) = XI(:, NN)
    XI(:, NN) = XIT(:)
  END DO
END SUBROUTINE POWELL
!
SUBROUTINE LINMIN(P, XI, N, FRET, TOL)
  Use Module_Data, ONLY : DP
  REAL(DP) :: FRET, P(N), XI(N), TOL, F1DIM
  REAL(DP) :: AX, BX, FA, FB, FX, XMIN, &
    XX, PCOM(2), XICOM(2), BRENT
  COMMON/F1COM/PCOM, XICOM, NCOM
  EXTERNAL F1DIM
  NCOM = 2
  PCOM = P
  XICOM = XI
  AX = 1._DP
  XX = 0._DP
  CALL MNBRAK(AX, XX, BX, FA, FX, FB, F1DIM)
  FRET = BRENT(AX, XX, BX, F1DIM, TOL, XMIN)
  XI(1 : N) = XMIN * XI(1 : N)
  P(1 : N) = P(1 : N) + XI(1 : N)
END SUBROUTINE LINMIN
!
SUBROUTINE MNBRAK(AX, BX, CX, FA, FB, FC, FUN)
  Use Module_Data, ONLY : DP 
  REAL(DP) :: AX, BX, CX, FA, FB, FC, FUN, &
    GOLD, GLIMIT, TINY, DUM, FU, Q, R, U, ULIM
  EXTERNAL FUN
  PARAMETER(GOLD = 1.618034_DP, &
    GLIMIT = 100._DP, TINY = 1.E-20_DP)
  FA = FUN(AX)
  FB = FUN(BX)
  IF(FB > FA)THEN
    DUM = AX
    AX = BX
    BX = DUM
    DUM = FB
    FB = FA
    FA = DUM
  END IF
  CX = BX + GOLD * (BX - AX)
  FC = FUN(CX)
  DO
    IF(FB<FC)RETURN
    R = (BX - AX) * (FB - FC)
    Q = (BX - CX) * (FB - FA)
    U = BX - ((BX - CX) * Q - (BX - AX) * R) / &
      (2._DP * SIGN(MAX(ABS(Q - R), TINY), Q - R))
    ULIM = BX + GLIMIT * (CX - BX)
    IF((BX - U) * (U - CX) > 0._DP)THEN
      FU = FUN(U)
      IF(FU<FC)THEN
        AX = BX
        FA = FB
        BX = U
        FB = FU
        RETURN
      ELSEIF(FU > FB)THEN
        CX = U
        FC = FU
        RETURN
      END IF
      U = CX + GOLD * (CX - BX)
      FU = FUN(U)
    ELSE IF((CX - U) * (U - ULIM) > 0._DP)THEN
      FU = FUN(U)
      IF(FU<FC)THEN
        BX = CX
        CX = U
        U = CX + GOLD * (CX - BX)
        FB = FC
        FC = FU
        FU = FUN(U)
      END IF
    ELSE IF((U - ULIM) * (ULIM - CX) >= 0._DP)THEN
      U = ULIM
      FU = FUN(U)
    ELSE
      U = CX + GOLD * (CX - BX)
      FU = FUN(U)
    END IF
    AX = BX
    BX = CX
    CX = U
    FA = FB
    FB = FC
    FC = FU
  END DO
END SUBROUTINE MNBRAK
!
FUNCTION F1DIM(X)
  Use Module_Data, ONLY : DP
  REAL(DP) :: F1DIM, FUNC, X, PCOM(2), XICOM(2), XT(2)
  COMMON/F1COM/PCOM, XICOM, NCOM
  XT(1 : NCOM) = PCOM(1 : NCOM) + X * XICOM(1 : NCOM)
  F1DIM = FUNC(XT, 2, NINT(XT(1)))
END FUNCTION F1DIM
!
FUNCTION BRENT(AX, BX, CX, F, TOL, XMIN)
  Use Module_Data, ONLY : DP, Thd0Q
  REAL(DP) :: BRENT, AX, BX, CX, TOL, XMIN, F, &
    CGOLD, ZEPS, A, B, D, E, ETEMP, FU, FV, FW, &
    FX, P, Q, R, TOL1, TOL2, U, V, W, X, XXMM, TINY
  PARAMETER(ITMAX = 100, CGOLD = 0.381966_DP, &
    ZEPS = 1.0E-10_DP, TINY = 1.E-100_DP)
  EXTERNAL F
  A = MIN(AX, CX)
  B = MAX(AX, CX)
  V = BX
  W = V
  X = V
  E = 0._DP
  FX = F(X)
  FV = FX
  FW = FX
  DO ITER = 1, ITMAX
    XXMM = 0.5_DP * (A + B)
    TOL1 = TOL * ABS(X) + ZEPS
    TOL2 = 2._DP * TOL1
    IF(ABS(X - XXMM) <= (TOL2 - 0.5_DP * (B - A)))GOTO 3
    IF(ABS(E) > TOL1)THEN
      R = (X - W) * (FX - FV)
      Q = (X - V) * (FX - FW)
      P = (X - V) * Q - (X - W) * R
      Q = 2._DP * (Q - R)
      IF(Q > 0._DP)P = -P
      Q = ABS(Q)
      ETEMP = E
      E = D
      IF(ABS(P) >= ABS(0.5_DP * Q * ETEMP) .OR. &
        P <= Q * (A - X) .OR. P >= Q * (B - X)) GOTO 1
      D = P / Q
      U = X + D
      IF(U - A < TOL2 .OR. B - U < TOL2)D = SIGN(TOL1, XXMM - X)
      GOTO 2
    END IF
1   IF(X >= XXMM)THEN
      E = A - X
    ELSE
      E = B - X
    END IF
    D = CGOLD * E
2     IF(ABS(D) >= TOL1)THEN
      U = X + D
    ELSE
      U = X + SIGN(TOL1, D)
    END IF
    FU = F(U)
    IF(FU <= FX)THEN
      IF(U >= X)THEN
        A = X
      ELSE
        B = X
      END IF
      V = W
      FV = FW
      W = X
      FW = FX
      X = U
      FX = FU
    ELSE
      IF(U<X)THEN
        A = U
      ELSE
        B = U
      END IF
      IF(FU <= FW .OR. ABS(W - X) < TINY) THEN
        V = W
        FV = FW
        W = U
        FW = FU
      ELSE IF(FU <= FV .OR. ABS(V - X) < TINY .OR. &
             ABS(V - W) < TINY) THEN
        V = U
        FV = FU
      END IF
    END IF
  END DO
  IF(Thd0Q) WRITE(*, *)'BRENT EXCEED MAXIMUM ITERATIONS'
3 XMIN = X
  BRENT = FX
END FUNCTION BRENT
!
SUBROUTINE BINOM(X, YLOG, XT, J, IDIM, NA1, NA2)
  Use Module_Data
  REAL(DP) :: X, YLOG, XT
  X = 0._DP
  YLOG = -1.E300_DP
  IF(J == 0)THEN
    IF(NA1 == 0)THEN
      X = 1._DP
      YLOG = 0._DP
    END IF
  ELSEIF(J == IDIM)THEN
    IF(NA2 == 0)THEN
      X = 1._DP
      YLOG = 0._DP
    END IF
  ELSE
    YLOG = XT + NA1 * (FAC(J) - FAC(J - 1)) + &
      NA2 * (FAC(IDIM - J) - FAC(IDIM - J - 1))
    IF(YLOG > -100._DP)X = EXP(YLOG)
  END IF
END SUBROUTINE BINOM
!
SUBROUTINE GTRANS(NR, XR, XM, IPA, MAXDIM)
  Use Module_Data
  REAL(DP) :: X1, X3, XMAX, XMAX2, XR, XM, &
    XM1, XM2, XM3, XM4, Y1, YLOG
  REAL(DP), Allocatable :: PTT(:)
  INTEGER :: LSTEP(2), LBOUNDS(2), LB(2), KB(2), &
    LB2, MAXDIM2
  PARAMETER(TINY = 1.E-9, BIG = 1. - TINY)
!$  INTEGER, PARAMETER :: CHUNK = 1
!$  INTEGER, EXTERNAL :: omp_get_thread_num
  DATA LSTEP/-1, 1/, LBOUNDS/0, 1/, NR_OLD/-1/
  SAVE NR_OLD
  Allocate(PTT(0 : NR))
  LBOUNDS(2) = NR
  XM1 = (1._DP - XM) * XR
  XM2 = IPA * XM / NINTPA
  XM3 = XM2 * NR
  XM4 = 1._DP - XM
  IF(ABS(XM4)<TINY)THEN
    XM4 = 1._DP
  ELSE
    XM4 = 1._DP / XM4
  END IF
  IF(NR /= NR_OLD) THEN
    IF(NR_OLD /= -1) DEALLOCATE(COL_LOG)
    ALLOCATE(COL_LOG(0 : NR, 0 : 3))
  END IF
  X1 = XM2 / PREC
  X3 = -XM1
  DO J = 0, NR
    X3 = X3 + XM1
    Y1 = X3 + XM2
    COL_LOG(J, 0) = Y1
    IF(Y1 > TINY .AND. Y1<BIG) THEN
      IF(NR /= NR_OLD .OR. X1 >= X3) THEN
        COL_LOG(J, 1) = LOG(Y1)
        COL_LOG(J, 2) = LOG(1. - Y1)
        COL_LOG(J, 3) = Y1 / (1. - Y1)
      END IF
    END IF
  END DO
  M = 0
! Open MP parallel for multiple processors
!$OMP PARALLEL IF(NumThread > 1) NUM_THREADS(NumThread) &
!$OMP   DEFAULT(SHARED) Private(XMAX, XMAX2, i2, X3, JJ, &
!$OMP   KB, LB, L, J, Y1, X1, YLOG, LB2, NN, JMAX, PTT)
!$OMP DO SCHEDULE(dynamic, CHUNK)
  DO I = 0, NR
    XMAX = -1._DP
    I2 = NR - I
    x3 = FAC(NR) - fac(i) - fac(I2)
    JJ = NINT((I - XM3) * XM4)
    IF(JJ <= 0)THEN
      JJ = 1
    ELSEIF(JJ >= NR)THEN
      JJ = NR - 1
    END IF
    KB(1) = JJ - 3
    IF(KB(1)<0)KB(1) = 0
    KB(2) = JJ + 3
    IF(KB(2) > NR)KB(2) = NR
    DO L = 1, 2
      LB(L) = LBOUNDS(L)
      DO j = JJ + L - 1, LBOUNDS(L), LSTEP(L)
        Y1 = COL_LOG(J, 0)
        X1 = 0._DP
        YLOG = -1.E300_DP
        IF(Y1 <= TINY)THEN
          IF(I == 0)THEN
            X1 = 1._DP
            YLOG = 0._DP
          END IF
        ELSEIF(Y1 >= BIG)THEN
          IF(I == NR)THEN
            X1 = 1._DP
            YLOG = 0._DP
          END IF
        ELSE
          YLOG = X3 + I * COL_LOG(J, 1) + I2 * COL_LOG(J, 2)
          IF(YLOG > -50) X1 = EXP(REAL(YLOG))
        END IF
        PTT(J) = X1
        IF(X1 > XMAX)THEN
          XMAX = X1
          XMAX2 = YLOG + PREC_LOG
          JMAX = J
        ELSEIF(YLOG < XMAX2)THEN
          LB(L) = J
          EXIT
        END IF
        IF(XMAX < TINY .AND. J == KB(L))THEN
          LB(L) = J
          EXIT
        END IF
      END DO
    END DO
    IF(XMAX >= TINY)THEN
      NN = LB(1)
      XMAX2 = XMAX * PREC
      LB2 = 0
      DO J = JMAX - 1, NN, -1
        IF(LB2 /= 0)CYCLE
        IF(PTT(J) < XMAX2)THEN
          LB(1) = J
          LB2 = 1
        END IF
      END DO
    ELSE
      LB = JMAX
    END IF
!$OMP CRITICAL
    IB(1, I) = M + 1
    DO J = LB(1), LB(2)
      IF(M + 1 > MAXDIM)THEN
        MAXDIM2 = INT(MAXDIM * 1.1)
        ALLOCATE(TRMX2(MAXDIM2), STAT = KK1)
        IF(KK1 == 0)THEN
          TRMX2(1 : M) = TRMX(1 : M)
          DEALLOCATE(TRMX)
          MAXDIM = MAXDIM2
          ALLOCATE(TRMX(MAXDIM), STAT = KK2)
          IF(KK2 == 0)THEN
            TRMX(1 : M) = TRMX2(1 : M)
            DEALLOCATE(TRMX2)
          END IF
        END IF
        IF(KK1 /= 0 .OR. KK2 /= 0)THEN
          IF(Thd0Q) WRITE(*, '(a/a/A)') &
            'The maximum value of Ne set ' // &
            'in your data is too large', &
            ' that your computer has not ' // &
            'enough RAM to run the program. ', &
            ' Try a smaller maximum value of ' // &
            'Ne and re-run the program'
          STOP
        END IF
      END IF
      M = M + 1
      TRMX(M) = PTT(J)
    END DO
    IB(2, I) = M
    IB(3, I) = LB(1)
!$OMP END CRITICAL
  END DO
!$OMP END DO NOWAIT
!$OMP END PARALLEL
  NR_OLD = NR
END SUBROUTINE GTRANS
!
FUNCTION RAND1(IDUM)
  INTEGER, PARAMETER :: SP = KIND(1.0)
  INTEGER IDUM, IM1, IM2, IMM1, IA1, &
    IA2, IQ1, IQ2, IR1, IR2, NTAB, NDIV
  REAL(SP) :: RAND1, AM, EPS, RNMX
  PARAMETER (IM1 = 2147483563, IM2 = 2147483399, &
    AM = 1. / IM1, IMM1 = IM1 - 1, IA1 = 40014, &
    IA2 = 40692, IQ1 = 53668, IQ2 = 52774, IR1 = 12211, &
    IR2 = 3791, NTAB = 32, NDIV = 1 + IMM1 / NTAB, &
    EPS = 1.2E-7, RNMX = 1. - EPS)
  INTEGER IDUM2, J, K, IV(NTAB), IY
  SAVE IV, IY, IDUM2
  DATA IDUM2/123456789/, IV/NTAB * 0/, IY/0/
  IF(IDUM <= 0)THEN
    IDUM = MAX(-IDUM, 1)
    IDUM2 = IDUM
    DO J = NTAB + 8, 1, -1
      K = IDUM / IQ1
      IDUM = IA1 * (IDUM - K * IQ1) - K * IR1
      IF(IDUM<0)IDUM = IDUM + IM1
      IF(J <= NTAB)IV(J) = IDUM
    END DO
    IY = IV(1)
  END IF
  K = IDUM / IQ1
  IDUM = IA1 * (IDUM - K * IQ1) - K * IR1
  IF(IDUM<0)IDUM = IDUM + IM1
  K = IDUM2 / IQ2
  IDUM2 = IA2 * (IDUM2 - K * IQ2) - K * IR2
  IF(IDUM2<0)IDUM2 = IDUM2 + IM2
  J = 1 + IY / NDIV
  IY = IV(J) - IDUM2
  IV(J) = IDUM
  IF(IY<1)IY = IY + IMM1
  RAND1 = MIN(AM * IY, RNMX)
END FUNCTION RAND1
!
SUBROUTINE polint(xa, ya, n, x, y, dy)
  Use Module_Data, ONLY : DP, Thd0Q 
  REAL(DP) :: dy, x, y, xa(n), ya(n), TINY
  PARAMETER (NMAX = 10, TINY = 1.E-50_DP)
  REAL(DP) :: den, dif, dift, ho, hp, w, c(NMAX), d(NMAX)
  ns = 1
  dif = abs(x - xa(1))
  do i = 1, n
    dift = abs(x - xa(i))
    if (dift<dif) then
      ns = i
      dif = dift
    END IF
    c(i) = ya(i)
    d(i) = ya(i)
  END DO
  y = ya(ns)
  ns = ns - 1
  do m = 1, n - 1
    do i = 1, n - m
      ho = xa(i) - x
      hp = xa(i + m) - x
      w = c(i + 1) - d(i)
      den = ho - hp
      if(abs(den) <= tiny)then
        IF(Thd0Q) write(*, *) &
          'Failure in polint, program stopped.'
        STOP
      END IF
      den = w / den
      d(i) = hp * den
      c(i) = ho * den
    END DO
    if (2 * ns<n - m)then
      dy = c(ns + 1)
    else
      dy = d(ns)
      ns = ns - 1
    END IF
    y = y + dy
  END DO
END SUBROUTINE polint

FUNCTION gammln(xx)
  INTEGER j
  DOUBLE PRECISION ser, stp, tmp, x, y, cof(6), gammln, xx
  SAVE cof, stp
  DATA cof, stp/76.18009172947146d0, -86.50532032941677d0, &
   24.01409824083091d0, -1.231739572450155d0, .1208650973866179d-2, &
   -.5395239384953d-5, 2.5066282746310005d0/
  x = xx
  y = x
  tmp = x + 5.5d0
  tmp = (x + 0.5d0) * log(tmp) - tmp
  ser = 1.000000000190015d0
  DO j = 1, 6
    y = y + 1.d0
    ser = ser + cof(j) / y
  END DO
  gammln = tmp + log(stp * ser / x)
END FUNCTION gammln

SUBROUTINE indexx(n,arr,indx)
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: n
  DOUBLE PRECISION, INTENT(IN) :: arr(n)
  INTEGER, INTENT(OUT) :: indx(n)
  INTEGER :: M, NSTACK
  DOUBLE PRECISION :: a
  PARAMETER (M = 7, NSTACK = 50)
  INTEGER :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  FORALL(j = 1 : n) indx(j) = j
  L = 1
  DO j = 2, n
    IF(arr(J) >= arr(J - 1)) CYCLE
    L = 0
    EXIT
  END DO
  IF(L == 1) RETURN
  jstack=0
  l=1
  ir=n
  do
    if(ir-l < M)then
      LoopJ : DO j=l+1,ir
        indxt=indx(j)
        a=arr(indxt)
        DO i=j-1,l,-1
          if(arr(indx(i)) <= a) then
            indx(i+1)=indxt
            CYCLE LoopJ
          endif
          indx(i+1)=indx(i)
        END DO
        i=l-1
        indx(i+1)=indxt
      END DO LoopJ
      if(jstack == 0)return
      ir=istack(jstack)
      l=istack(jstack-1)
      jstack=jstack-2
    else
      k=(l+ir)/2
      itemp=indx(k)
      indx(k)=indx(l+1)
      indx(l+1)=itemp
      if(arr(indx(l)) > arr(indx(ir)))then
        itemp=indx(l)
        indx(l)=indx(ir)
        indx(ir)=itemp
      endif
      if(arr(indx(l+1)) > arr(indx(ir)))then
        itemp=indx(l+1)
        indx(l+1)=indx(ir)
        indx(ir)=itemp
      endif
      if(arr(indx(l)) > arr(indx(l+1)))then
        itemp=indx(l)
        indx(l)=indx(l+1)
        indx(l+1)=itemp
      endif
      i=l+1
      j=ir
      indxt=indx(l+1)
      a=arr(indxt)
      do
        i=i+1
        if(arr(indx(i)) < a) cycle
        do
          j=j-1
          if(arr(indx(j)) <= a) exit
        end do
        if(j < i) exit
        itemp=indx(i)
        indx(i)=indx(j)
        indx(j)=itemp
      enddo
      indx(l+1)=indx(j)
      indx(j)=indxt
      jstack=jstack+2
      if(jstack > NSTACK) THEN
        WRITE(*, '(A)') 'Program stops due to ' // &
          'NSTACK too small in subroutine Indexx'
        STOP
      END IF
      if(ir-i+1 >= j-l)then
        istack(jstack)=ir
        istack(jstack-1)=i
        ir=j-1
      else
        istack(jstack)=j-1
        istack(jstack-1)=l
        l=i
      endif
    endif
  enddo
END SUBROUTINE indexx